#! /usr/bin/env python
from __future__ import print_function
from __future__ import absolute_import
import os
import sys
import unittest
import time
import calendar

import irods.test.helpers as helpers
import tempfile
from irods.session import iRODSSession
import irods.exception as ex
import irods.keywords as kw
from irods.ticket import Ticket
from irods.models import (TicketQuery,DataObject,Collection)


# As with most of the modules in this test suite, session objects created via
# make_session() are implicitly agents of a rodsadmin unless otherwise indicated.
# Counterexamples within this module shall be obvious as they are instantiated by
# the login() method, and always tied to one of the traditional rodsuser names
# widely used in iRODS test suites, ie. 'alice' or 'bob'.


def gmtime_to_timestamp (gmt_struct):
    return "{0.tm_year:04d}-{0.tm_mon:02d}-{0.tm_mday:02d}."\
           "{0.tm_hour:02d}:{0.tm_min:02d}:{0.tm_sec:02d}".format(gmt_struct)


def delete_my_tickets(session):
    my_userid = session.users.get( session.username ).id
    my_tickets = session.query(TicketQuery.Ticket).filter(TicketQuery.Ticket.user_id ==  my_userid)
    for res in my_tickets:
        Ticket(session, result = res).delete()


class TestRodsUserTicketOps(unittest.TestCase):

    def login(self,user):
        return iRODSSession (port=self.port,zone=self.zone,host=self.host,
                user=user.name,password=self.users[user.name])

    @staticmethod
    def irods_homedir(sess, path_only = False):
        path = '/{0.zone}/home/{0.username}'.format(sess)
        if path_only:
            return path
        return sess.collections.get(path)

    @staticmethod
    def list_objects (sess):
        return [ '{}/{}'.format(o[Collection.name],o[DataObject.name]) for o in
            sess.query(Collection.name,DataObject.name) ]

    users = {
            'alice':'apass',
            'bob':'bpass'
            }

    def setUp(self):

        self.alice = self.bob = None

        with helpers.make_session() as ses:
            u = ses.users.get(ses.username)
            if u.type != 'rodsadmin':
                self.skipTest('''Test runnable only by rodsadmin.''')
            self.host = ses.host
            self.port = ses.port
            self.zone = ses.zone
            for newuser,passwd in self.users.items():
                u = ses.users.create( newuser, 'rodsuser')
                setattr(self,newuser,u)
                u.modify('password', passwd)

    def tearDown(self):
        with helpers.make_session() as ses:
            for u in self.users:
                ses.users.remove(u)


    def test_admin_keyword_for_tickets (self):

        N_TICKETS = 3

        # Create some tickets as alice.

        with self.login(self.alice) as alice:
            alice_home_path = self.irods_homedir(alice, path_only = True)
            ticket_strings = [ Ticket(alice).issue('read', alice_home_path).string for _ in range(N_TICKETS) ]

        # As rodsadmin, use the ADMIN_KW flag to delete alice's tickets.

        with helpers.make_session() as ses:
            alices_tickets = [t[TicketQuery.Ticket.string] for t in ses.query(TicketQuery.Ticket).filter(TicketQuery.Owner.name == 'alice')]
            self.assertEqual(len(alices_tickets),N_TICKETS)
            for s in alices_tickets:
                Ticket( ses, s ).delete(**{kw.ADMIN_KW:''})
            alices_tickets = [t[TicketQuery.Ticket.string] for t in ses.query(TicketQuery.Ticket).filter(TicketQuery.Owner.name == 'alice')]
            self.assertEqual(len(alices_tickets),0)


    def test_ticket_expiry (self):
        with helpers.make_session() as ses:
            t1 = t2 = dobj = None
            try:
                gm_now = time.gmtime()
                gm_later = time.gmtime( calendar.timegm( gm_now ) + 10 )
                home = self.irods_homedir(ses)
                dobj = helpers.make_object(ses, home.path+'/dummy', content='abcxyz')

                later_ts = gmtime_to_timestamp (gm_later)
                later_epoch = calendar.timegm (gm_later)

                t1 = Ticket(ses)
                t2 = Ticket(ses)

                tickets = [ _.issue('read',dobj.path).string for _ in (t1,
                                                                       t2,) ]
                t1.modify('expire',later_ts)    # - Specify expiry with the human readable timestamp.
                t2.modify('expire',later_epoch) # - Specify expiry formatted as epoch seconds.

                # Check normal access succeeds prior to expiration
                for ticket_string in tickets:
                    with self.login(self.alice) as alice:
                        Ticket(alice, ticket_string).supply()
                        alice.data_objects.get(dobj.path)

                # Check that both time formats have effected the same expiry time (The catalog returns epoch secs.)
                timestamps = []
                for ticket_string in tickets:
                    t = ses.query(TicketQuery.Ticket).filter(TicketQuery.Ticket.string == ticket_string).one()
                    timestamps.append( t [TicketQuery.Ticket.expiry_ts] )
                self.assertEqual (len(timestamps),2)
                self.assertEqual (timestamps[0],timestamps[1])

                # Wait for tickets to expire.
                epoch = int(time.time())
                while epoch <= later_epoch:
                    time.sleep(later_epoch - epoch + 1)
                    epoch = int(time.time())

                Expected_Exception = ex.CAT_TICKET_EXPIRED if ses.server_version >= (4,2,9) \
                        else ex.SYS_FILE_DESC_OUT_OF_RANGE

                # Check tickets no longer allow access.
                for ticket_string in tickets:
                    with self.login(self.alice) as alice, tempfile.NamedTemporaryFile() as f:
                        Ticket(alice, ticket_string).supply()
                        with self.assertRaises( Expected_Exception ):
                            alice.data_objects.get(dobj.path,f.name, **{kw.FORCE_FLAG_KW:''})

            finally:
                if t1: t1.delete()
                if t2: t2.delete()
                if dobj: dobj.unlink(force=True)


    def test_object_read_and_write_tickets(self):
        if self.alice is None or self.bob is None:
            self.skipTest("A rodsuser (alice and/or bob) could not be created.")
        t=None
        data_objs=[]
        tmpfiles=[]
        try:
            # Create ticket for read access to alice's home collection.
            alice = self.login(self.alice)
            home = self.irods_homedir(alice)

            # Create 'R' and 'W' in alice's home collection.
            data_objs = [helpers.make_object(alice,home.path+"/"+name,content='abcxyz') for name in ('R','W')]
            tickets = {
                'R': Ticket(alice).issue('read',  home.path + "/R").string,
                'W': Ticket(alice).issue('write', home.path + "/W").string
            }
            # Test only write ticket allows upload.
            with self.login(self.bob) as bob:
                rw_names={}
                for name in  ('R','W'):
                    Ticket( bob, tickets[name] ).supply()
                    with tempfile.NamedTemporaryFile (delete=False) as tmpf:
                        tmpfiles += [tmpf]
                        rw_names[name] = tmpf.name
                        tmpf.write(b'hello')
                    if name=='W':
                        bob.data_objects.put(tmpf.name,home.path+"/"+name)
                    else:
                        try:
                            bob.data_objects.put(tmpf.name,home.path+"/"+name)
                        except ex.CAT_NO_ACCESS_PERMISSION:
                            pass
                        else:
                            raise AssertionError("A read ticket allowed a data object write operation to happen without error.")

            # Test upload was successful, by getting and confirming contents.

            with self.login(self.bob) as bob:  # This check must be in a new session or we get CollectionDoesNotExist. - Possibly a new issue [ ]
                for name in  ('R','W'):
                    Ticket( bob, tickets[name] ).supply()
                    bob.data_objects.get(home.path+"/"+name,rw_names[ name ],**{kw.FORCE_FLAG_KW:''})
                    with open(rw_names[ name ],'r') as tmpread:
                        self.assertEqual(tmpread.read(),
                                         'abcxyz' if name == 'R' else 'hello')
        finally:
            if t: t.delete()
            for d in data_objs:
                d.unlink(force=True)
            for file_ in tmpfiles: os.unlink( file_.name )
            alice.cleanup()


    def test_coll_read_ticket_between_rodsusers(self):
        t=None
        data_objs=[]
        tmpfiles=[]
        try:
            # Create ticket for read access to alice's home collection.
            alice = self.login(self.alice)
            tc = Ticket(alice)
            home = self.irods_homedir(alice)
            tc.issue('read', home.path)

            # Create 'x' and 'y' in alice's home collection
            data_objs = [helpers.make_object(alice,home.path+"/"+name,content='abcxyz') for name in ('x','y')]

            with self.login(self.bob) as bob:
                ts = Ticket( bob, tc.string )
                ts.supply()
                # Check collection access ticket allows bob to list both subobjects
                self.assertEqual(len(self.list_objects(bob)),2)
                # and that we can get (and read) them properly.
                for name in ('x','y'):
                    with tempfile.NamedTemporaryFile (delete=False) as tmpf:
                        tmpfiles += [tmpf]
                    bob.data_objects.get(home.path+"/"+name,tmpf.name,**{kw.FORCE_FLAG_KW:''})
                    with open(tmpf.name,'r') as tmpread:
                        self.assertEqual(tmpread.read(),'abcxyz')

            td = Ticket(alice)
            td.issue('read', home.path+"/x")

            with self.login(self.bob) as bob:
                ts = Ticket( bob, td.string )
                ts.supply()

                # Check data access ticket allows bob to list only one data object
                self.assertEqual(len(self.list_objects(bob)),1)

                # ... and fetch that object (verifying content)
                with tempfile.NamedTemporaryFile (delete=False) as tmpf:
                    tmpfiles += [tmpf]
                bob.data_objects.get(home.path+"/x",tmpf.name,**{kw.FORCE_FLAG_KW:''})
                with open(tmpf.name,'r') as tmpread:
                    self.assertEqual(tmpread.read(),'abcxyz')

                # ... but not fetch the other data object owned by alice.
                with self.assertRaises(ex.DataObjectDoesNotExist):
                    bob.data_objects.get(home.path+"/y")
        finally:
            if t: t.delete()
            for d in data_objs:
                d.unlink(force=True)
            for file_ in tmpfiles: os.unlink( file_.name )
            alice.cleanup()


class TestTicketOps(unittest.TestCase):

    def setUp(self):
        """Create objects for test"""
        self.sess = helpers.make_session()
        user = self.sess.users.get(self.sess.username)
        if user.type != 'rodsadmin':
            self.skipTest('''Test runnable only by rodsadmin.''')

        admin = self.sess
        delete_my_tickets( admin )

        # Create test collection

        self.coll_path = '/{}/home/{}/ticket_test_dir'.format(admin.zone, admin.username)
        self.coll = helpers.make_collection(admin, self.coll_path)

        # Create anonymous test user
        self.user = admin.users.create('anonymous','rodsuser')
        self.rodsuser_params = { 'host':admin.host,
                                 'port':admin.port,
                                 'user': 'anonymous',
                                 'password':'',
                                 'zone':admin.zone }

        # make new data object in the test collection with some initialized content

        self.INITIALIZED_DATA = b'1'*16
        self.data_path = '{self.coll_path}/ticketed_data'.format(**locals())
        helpers.make_object (admin, self.data_path, content = self.INITIALIZED_DATA)

        self.MODIFIED_DATA = b'2'*16

        # make new tickets for the various combinations

        self.tickets = {'coll':{},'data':{}}
        for obj_type in ('coll','data'):
            for access in ('read','write'):
                ticket = Ticket(admin)
                self.tickets [obj_type] [access] = ticket.string
                ticket.issue( access , getattr(self, obj_type + '_path'))


    def tearDown(self):
        """Clean up tickets , collections and data objects used for test."""
        admin = self.sess
        delete_my_tickets( admin )
        if getattr(self,'coll',None):
            self.coll.remove(recurse=True, force=True)
        if getattr(self,'user',None):
            self.user.remove()
        admin.cleanup()


    def _ticket_read_helper( self, obj_type, download = False ):
        with iRODSSession( ** self.rodsuser_params ) as user_sess:
            temp_file = []
            if download: temp_file += [tempfile.mktemp()]
            try:
                Ticket(user_sess, self.tickets[obj_type]['read']).supply()
                data = user_sess.data_objects.get(self.data_path,*temp_file)
                self.assertEqual (data.open('r').read(), self.INITIALIZED_DATA)
                if temp_file:
                    with open(temp_file[0],'rb') as local_file:
                        self.assertEqual (local_file.read(), self.INITIALIZED_DATA)
            finally:
                if temp_file and os.path.exists(temp_file[0]):
                    os.unlink(temp_file[0])


    def test_data_ticket_read(self): self._ticket_read_helper( obj_type = 'data' )

    def test_coll_ticket_read(self): self._ticket_read_helper( obj_type = 'coll' )

    def test_data_ticket_read_with_download(self): self._ticket_read_helper( obj_type = 'data', download = True )

    def test_coll_ticket_read_with_download(self): self._ticket_read_helper( obj_type = 'coll', download = True )


    def _ticket_write_helper( self, obj_type ):
        with iRODSSession( ** self.rodsuser_params ) as user_sess:
            Ticket(user_sess, self.tickets[obj_type]['write']).supply()
            data = user_sess.data_objects.get(self.data_path)
            with data.open('w') as obj:
                obj.write(self.MODIFIED_DATA)
            self.assertEqual (data.open('r').read(), self.MODIFIED_DATA)


    def test_data_ticket_write(self): self._ticket_write_helper( obj_type = 'data' )

    def test_coll_ticket_write(self): self._ticket_write_helper( obj_type = 'coll' )


if __name__ == '__main__':
    # let the tests find the parent irods lib
    sys.path.insert(0, os.path.abspath('../..'))
    unittest.main()



#! /usr/bin/env python
from __future__ import absolute_import
import os
import sys
import unittest
from irods.exception import NetworkException
import irods.test.helpers as helpers


class TestConnections(unittest.TestCase):

    def setUp(self):
        self.sess = helpers.make_session()

    def tearDown(self):
        '''Close connections
        '''
        self.sess.cleanup()

    def test_connection(self):
        with self.sess.pool.get_connection() as conn:
            self.assertTrue(conn)

    def test_connection_destructor(self):
        conn = self.sess.pool.get_connection()
        conn.__del__()
        conn.release(destroy=True)

    def test_failed_connection(self):
        # mess with the account's port
        saved_port = self.sess.port
        self.sess.pool.account.port = 6666

        # try connecting
        with self.assertRaises(NetworkException):
            self.sess.pool.get_connection()

        # set port back
        self.sess.pool.account.port = saved_port

    def test_send_failure(self):
        with self.sess.pool.get_connection() as conn:
            # try to close connection twice, 2nd one should fail
            conn.disconnect()
            with self.assertRaises(NetworkException):
                conn.disconnect()

    def test_reply_failure(self):
        with self.sess.pool.get_connection() as conn:
            # close connection
            conn.disconnect()

            # try sending reply
            with self.assertRaises(NetworkException):
                conn.reply(0)


if __name__ == '__main__':
    # let the tests find the parent irods lib
    sys.path.insert(0, os.path.abspath('../..'))
    unittest.main()



from __future__ import absolute_import
import datetime
import logging
import threading
import os

from irods import DEFAULT_CONNECTION_TIMEOUT
from irods.connection import Connection

logger = logging.getLogger(__name__)

def attribute_from_return_value(attrname):
    def deco(method):
        def method_(self,*s,**kw):
            ret = method(self,*s,**kw)
            setattr(self,attrname,ret)
            return ret
        return method_
    return deco

DEFAULT_APPLICATION_NAME = 'python-irodsclient'

class Pool(object):

    def __init__(self, account, application_name='', connection_refresh_time=-1):
        '''
        Pool( account , application_name='' )
        Create an iRODS connection pool; 'account' is an irods.account.iRODSAccount instance and
        'application_name' specifies the application name as it should appear in an 'ips' listing.
        '''

        self._thread_local = threading.local()
        self.account = account
        self._lock = threading.RLock()
        self.active = set()
        self.idle = set()
        self.connection_timeout = DEFAULT_CONNECTION_TIMEOUT
        self.application_name = ( os.environ.get('spOption','') or
                                  application_name or
                                  DEFAULT_APPLICATION_NAME )

        if connection_refresh_time > 0:
            self.refresh_connection = True
            self.connection_refresh_time = connection_refresh_time
        else:
            self.refresh_connection = False
            self.connection_refresh_time = None

    @property
    def _conn(self): return getattr( self._thread_local, "_conn", None)

    @_conn.setter
    def _conn(self, conn_): setattr( self._thread_local, "_conn", conn_)

    @attribute_from_return_value("_conn")
    def get_connection(self):
        with self._lock:
            try:
                conn = self.idle.pop()

                curr_time = datetime.datetime.now()
                # If 'refresh_connection' flag is True and the connection was
                # created more than 'connection_refresh_time' seconds ago,
                # release the connection (as its stale) and create a new one
                if self.refresh_connection and (curr_time - conn.create_time).total_seconds() > self.connection_refresh_time:
                    logger.debug('Connection with id {} was created more than {} seconds ago. Releasing the connection and creating a new one.'.format(id(conn), self.connection_refresh_time))
                    self.release_connection(conn, True)
                    conn = Connection(self, self.account)
                    logger.debug("Created new connection with id: {}".format(id(conn)))
            except KeyError:
                conn = Connection(self, self.account)
                logger.debug("No connection found in idle set. Created a new connection with id: {}".format(id(conn)))

            self.active.add(conn)
            logger.debug("Adding connection with id {} to active set".format(id(conn)))

        logger.debug('num active: {}'.format(len(self.active)))
        logger.debug('num idle: {}'.format(len(self.idle)))
        return conn

    def release_connection(self, conn, destroy=False):
        with self._lock:
            if conn in self.active:
                self.active.remove(conn)
                logger.debug("Removed connection with id: {} from active set".format(id(conn)))
                if not destroy:
                    # If 'refresh_connection' flag is True, update connection's 'last_used_time'
                    if self.refresh_connection:
                        conn.last_used_time = datetime.datetime.now()
                    self.idle.add(conn)
                    logger.debug("Added connection with id: {} to idle set".format(id(conn)))
            elif conn in self.idle and destroy:
                logger.debug("Destroyed connection with id: {}".format(id(conn)))
                self.idle.remove(conn)
        logger.debug('num active: {}'.format(len(self.active)))
        logger.debug('num idle: {}'.format(len(self.idle)))



from __future__ import absolute_import
from irods.column import Column, Integer, String, DateTime, Keyword
import six


class ModelBase(type):
    columns = {}

    def __new__(cls, name, bases, attr):
        columns = [y for (x, y) in six.iteritems(attr) if isinstance(y, Column)]
        for col in columns:
            ModelBase.columns[col.icat_id] = col
        attr['_columns'] = columns
        # attr['_icat_column_names'] = [y.icat_key for (x,y) in columns]
        return type.__new__(cls, name, bases, attr)


class Model(six.with_metaclass(ModelBase, object)):
    pass


class RuleExec(Model):
    id = Column(Integer, 'RULE_EXEC_ID', 1000)
    name = Column(String, 'RULE_EXEC_NAME', 1001)
    rei_file_path = Column(String,'RULE_EXEC_REI_FILE_PATH', 1002)
    user_name = Column(String, 'RULE_EXEC_USER_NAME', 1003)
    time = Column(DateTime,'RULE_EXEC_TIME',    1005)
    last_exe_time = Column(DateTime,'RULE_EXEC_LAST_EXE_TIME', 1010)
    frequency = Column(String,'RULE_EXEC_FREQUENCY', 1006)
    priority = Column(String, 'RULE_EXEC_PRIORITY', 1007)

#   # If needed in 4.2.9, we can update the Query class to dynamically
#   #  attach this field based on server version:
#   context = Column(String, 'RULE_EXEC_CONTEXT', 1012)

#   # These are either unused or usually absent:
#   exec_status = Column(String,'RULE_EXEC_STATUS', 1011)
#   address = Column(String,'RULE_EXEC_ADDRESS', 1004)
#   notification_addr = Column('RULE_EXEC_NOTIFICATION_ADDR', 1009)


class Zone(Model):
    id = Column(Integer, 'ZONE_ID', 101)
    name = Column(String, 'ZONE_NAME', 102)
    type = Column(String, 'ZONE_TYPE', 103)


class User(Model):
    id = Column(Integer, 'USER_ID', 201)
    name = Column(String, 'USER_NAME', 202)
    type = Column(String, 'USER_TYPE', 203)
    zone = Column(String, 'USER_ZONE', 204)
    info = Column(String, 'USER_INFO', 206)
    comment = Column(String, 'USER_COMMENT', 207)
    create_time = Column(DateTime, 'USER_CREATE_TIME', 208)
    modify_time = Column(DateTime, 'USER_MODIFY_TIME', 209)


class UserAuth(Model):
    user_id = Column(Integer, 'USER_AUTH_ID', 1600)
    user_dn = Column(String, 'USER_DN', 1601)


class CollectionUser(Model):
    name = Column(String, 'COL_COLL_USER_NAME', 1300)
    zone = Column(String, 'COL_COLL_USER_ZONE', 1301)


class UserGroup(Model):
    id = Column(Integer, 'USER_GROUP_ID', 900)
    name = Column(String, 'USER_GROUP_NAME', 901)


class Resource(Model):
    id = Column(Integer, 'R_RESC_ID', 301)
    name = Column(String, 'R_RESC_NAME', 302)
    zone_name = Column(String, 'R_ZONE_NAME', 303)
    type = Column(String, 'R_TYPE_NAME', 304)
    class_name = Column(String, 'R_CLASS_NAME', 305)
    location = Column(String, 'R_LOC', 306)
    vault_path = Column(String, 'R_VAULT_PATH', 307)
    free_space = Column(String, 'R_FREE_SPACE', 308)
    free_space_time = Column(String, 'R_FREE_SPACE_TIME', 314, min_version=(4,0,0))
    comment = Column(String, 'R_RESC_COMMENT', 310)
    create_time = Column(DateTime, 'R_CREATE_TIME', 311)
    modify_time = Column(DateTime, 'R_MODIFY_TIME', 312)
    status = Column(String, 'R_RESC_STATUS', 313, min_version=(4,0,0))
    children = Column(String, 'R_RESC_CHILDREN', 315, min_version=(4,0,0))
    context = Column(String, 'R_RESC_CONTEXT', 316, min_version=(4,0,0))
    parent = Column(String, 'R_RESC_PARENT', 317, min_version=(4,0,0))
    parent_context = Column(String, 'R_RESC_PARENT_CONTEXT', 318, min_version=(4,2,0))


class DataObject(Model):
    id = Column(Integer, 'D_DATA_ID', 401)
    collection_id = Column(Integer, 'D_COLL_ID', 402)
    name = Column(String, 'DATA_NAME', 403)  # basename
    replica_number = Column(Integer, 'DATA_REPL_NUM', 404)
    version = Column(String, 'DATA_VERSION', 405)
    type = Column(String, 'DATA_TYPE_NAME', 406)
    size = Column(Integer, 'DATA_SIZE', 407)
    resource_name = Column(String, 'D_RESC_NAME', 409)
    path = Column(String, 'D_DATA_PATH', 410)  # physical path on resource
    owner_name = Column(String, 'D_OWNER_NAME', 411)
    owner_zone = Column(String, 'D_OWNER_ZONE', 412)
    replica_status = Column(String, 'D_REPL_STATUS', 413)
    status = Column(String, 'D_DATA_STATUS', 414)
    checksum = Column(String, 'D_DATA_CHECKSUM', 415)
    expiry = Column(String, 'D_EXPIRY', 416)
    map_id = Column(Integer, 'D_MAP_ID', 417)
    comments = Column(String, 'D_COMMENTS', 418)
    create_time = Column(DateTime, 'D_CREATE_TIME', 419)
    modify_time = Column(DateTime, 'D_MODIFY_TIME', 420)
    resc_hier = Column(String, 'D_RESC_HIER', 422, min_version=(4,0,0))
    resc_id = Column(String, 'D_RESC_ID', 423, min_version=(4,2,0))


class Collection(Model):
    id = Column(Integer, 'COLL_ID', 500)
    name = Column(String, 'COLL_NAME', 501)
    parent_name = Column(String, 'COLL_PARENT_NAME', 502)
    owner_name = Column(String, 'COLL_OWNER_NAME', 503)
    owner_zone = Column(String, 'COLL_OWNER_ZONE', 504)
    map_id = Column(String, 'COLL_MAP_ID', 505)
    inheritance = Column(String, 'COLL_INHERITANCE', 506)
    comments = Column(String, 'COLL_COMMENTS', 507)
    create_time = Column(DateTime, 'COLL_CREATE_TIME', 508)
    modify_time = Column(DateTime, 'COLL_MODIFY_TIME', 509)


class DataObjectMeta(Model):
    id = Column(String, 'COL_META_DATA_ATTR_ID', 603)
    name = Column(String, 'COL_META_DATA_ATTR_NAME', 600)
    value = Column(String, 'COL_META_DATA_ATTR_VALUE', 601)
    units = Column(String, 'COL_META_DATA_ATTR_UNITS', 602)
    create_time = Column(DateTime, 'COL_META_DATA_CREATE_TIME', 604)
    modify_time = Column(DateTime, 'COL_META_DATA_MODIFY_TIME', 605)


class CollectionMeta(Model):
    id = Column(String, 'COL_META_COLL_ATTR_UNITS', 613)
    name = Column(String, 'COL_META_COLL_ATTR_NAME', 610)
    value = Column(String, 'COL_META_COLL_ATTR_VALUE', 611)
    units = Column(String, 'COL_META_COLL_ATTR_UNITS', 612)
    create_time = Column(DateTime, 'COL_META_COLL_CREATE_TIME', 614)
    modify_time = Column(DateTime, 'COL_META_COLL_MODIFY_TIME', 615)



class ResourceMeta(Model):
    id = Column(String, 'COL_META_RESC_ATTR_UNITS', 633)
    name = Column(String, 'COL_META_RESC_ATTR_NAME', 630)
    value = Column(String, 'COL_META_RESC_ATTR_VALUE', 631)
    units = Column(String, 'COL_META_RESC_ATTR_UNITS', 632)
    create_time = Column(DateTime, 'COL_META_RESC_CREATE_TIME', 634)
    modify_time = Column(DateTime, 'COL_META_RESC_MODIFY_TIME', 635)



class UserMeta(Model):
    id = Column(String, 'COL_META_USER_ATTR_ID', 643)
    name = Column(String, 'COL_META_USER_ATTR_NAME', 640)
    value = Column(String, 'COL_META_USER_ATTR_VALUE', 641)
    units = Column(String, 'COL_META_USER_ATTR_UNITS', 642)
    create_time = Column(DateTime, 'COL_META_USER_CREATE_TIME', 644)
    modify_time = Column(DateTime, 'COL_META_USER_MODIFY_TIME', 645)



class DataAccess(Model):
    type = Column(Integer, 'DATA_ACCESS_TYPE', 700)
    name = Column(String, 'COL_DATA_ACCESS_NAME', 701)
    token_namespace = Column(String, 'COL_DATA_TOKEN_NAMESPACE', 702)
    user_id = Column(Integer, 'COL_DATA_ACCESS_USER_ID', 703)
    data_id = Column(Integer, 'COL_DATA_ACCESS_DATA_ID', 704)


class CollectionAccess(Model):
    type = Column(Integer, 'COL_COLL_ACCESS_TYPE', 710)
    name = Column(String, 'COL_COLL_ACCESS_NAME', 711)
    token_namespace = Column(String, 'COL_COLL_TOKEN_NAMESPACE', 712)
    user_id = Column(Integer, 'COL_COLL_ACCESS_USER_ID', 713)
    access_id = Column(Integer, 'COL_COLL_ACCESS_COLL_ID', 714)


class SpecificQueryResult(Model):
    '''
    To parse results of specific queries. No corresponding iCAT table.
    '''
    value = Column(String, 'SQL_RESULT_VALUE', 0)


# not really a model. Should be dict instead?
class Keywords(Model):
    data_type = Keyword(String, 'dataType')
    chksum = Keyword(String, 'chksum')


class TicketQuery:
    """Various model classes for querying attributes of iRODS tickets.

    Namespacing these model classes under the TicketQuery parent class allows
    a simple import (not conflicting with irods.ticket.Ticket) and a usage
    that reflects ICAT table structure:

        from irods.models import TicketQuery
        # ...
        for row in session.query( TicketQuery.Ticket )\
                          .filter( TicketQuery.Owner.name == 'alice' ):
            print( row [TicketQuery.Ticket.string] )

    (For more examples, see irods/test/ticket_test.py)

    """
    class Ticket(Model):
        """For queries of R_TICKET_MAIN."""
        id = Column(Integer, 'TICKET_ID', 2200)
        string = Column(String, 'TICKET_STRING', 2201)
        type = Column(String, 'TICKET_TYPE', 2202)
        user_id = Column(Integer, 'TICKET_USER_ID', 2203)
        object_id = Column(Integer, 'TICKET_OBJECT_ID', 2204)
        object_type = Column(String, 'TICKET_OBJECT_TYPE', 2205)
        uses_limit = Column(Integer, 'TICKET_USES_LIMIT', 2206)
        uses_count = Column(Integer, 'TICKET_USES_COUNT', 2207)
        expiry_ts = Column(String, 'TICKET_EXPIRY_TS', 2208)
        write_file_count = Column(Integer, 'TICKET_WRITE_FILE_COUNT', 2211)
        write_file_limit = Column(Integer, 'TICKET_WRITE_FILE_LIMIT', 2212)
        write_byte_count = Column(Integer, 'TICKET_WRITE_BYTE_COUNT', 2213)
        write_byte_limit = Column(Integer, 'TICKET_WRITE_BYTE_LIMIT', 2214)
## For now, use of these columns raises CAT_SQL_ERR in both PRC and iquest: (irods/irods#5929)
#       create_time = Column(String, 'TICKET_CREATE_TIME', 2209)
#       modify_time = Column(String, 'TICKET_MODIFY_TIME', 2210)

    class DataObject(Model):
        """For queries of R_DATA_MAIN when joining to R_TICKET_MAIN.

        The ticket(s) in question should be for a data object; otherwise
        CAT_SQL_ERR is thrown.

        """
        name = Column(String, 'TICKET_DATA_NAME', 2226)
        coll = Column(String, 'TICKET_DATA_COLL_NAME', 2227)

    class Collection(Model):
        """For queries of R_COLL_MAIN when joining to R_TICKET_MAIN.

        The returned ticket(s) will be limited to those issued for collections.

        """
        name = Column(String, 'TICKET_COLL_NAME', 2228)

    class Owner(Model):
        """For queries concerning R_TICKET_USER_MAIN."""
        name = Column(String, 'TICKET_OWNER_NAME', 2229)
        zone = Column(String, 'TICKET_OWNER_ZONE', 2230)

    class AllowedHosts(Model):
        """For queries concerning R_TICKET_ALLOWED_HOSTS."""
        ticket_id = Column(String, 'COL_TICKET_ALLOWED_HOST_TICKET_ID', 2220)
        host = Column(String, 'COL_TICKET_ALLOWED_HOST', 2221)

    class AllowedUsers(Model):
        """For queries concerning R_TICKET_ALLOWED_USERS."""
        ticket_id = Column(String, 'COL_TICKET_ALLOWED_USER_TICKET_ID', 2222)
        user_name = Column(String, 'COL_TICKET_ALLOWED_USER', 2223)

    class AllowedGroups(Model):
        """For queries concerning R_TICKET_ALLOWED_GROUPS."""
        ticket_id = Column(String, 'COL_TICKET_ALLOWED_GROUP_TICKET_ID', 2224)
        group_name = Column(String, 'COL_TICKET_ALLOWED_GROUP', 2225)



from __future__ import absolute_import
from collections import OrderedDict

from irods import MAX_SQL_ROWS
from irods.models import Model
from irods.column import Column, Keyword
from irods.message import (
    IntegerIntegerMap, IntegerStringMap, StringStringMap, _OrderedMultiMapping,
    GenQueryRequest, GenQueryResponse, empty_gen_query_out,
    iRODSMessage, SpecificQueryRequest, GeneralAdminRequest)
from irods.api_number import api_number
from irods.exception import CAT_NO_ROWS_FOUND, MultipleResultsFound, NoResultFound
from irods.results import ResultSet, SpecificQueryResultSet
import six

query_number = {'ORDER_BY': 0x400,
                'ORDER_BY_DESC': 0x800,
                'RETURN_TOTAL_ROW_COUNT': 0x20,
                'NO_DISTINCT': 0x40,
                'QUOTA_QUERY': 0x80,
                'AUTO_CLOSE': 0x100,
                'UPPER_CASE_WHERE': 0x200,
                'SELECT_MIN': 2,
                'SELECT_MAX': 3,
                'SELECT_SUM': 4,
                'SELECT_AVG': 5,
                'SELECT_COUNT': 6}


class Query(object):

    def __init__(self, sess, *args, **kwargs):
        self.sess = sess
        self.columns = OrderedDict()
        self.criteria = []
        self._limit = -1
        self._offset = 0
        self._continue_index = 0
        self._keywords = {}

        for arg in args:
            if isinstance(arg, type) and issubclass(arg, Model):
                for col in arg._columns:
                    if self.sess.server_version >= col.min_version:
                        self.columns[col] = 1
            elif isinstance(arg, Column):
                self.columns[arg] = 1
            else:
                raise TypeError("Arguments must be models or columns")

    def _clone(self):
        new_q = Query(self.sess)
        new_q.columns = self.columns
        new_q.criteria = self.criteria
        new_q._limit = self._limit
        new_q._offset = self._offset
        new_q._continue_index = self._continue_index
        new_q._keywords = self._keywords
        return new_q

    def add_keyword(self, keyword, value = ''):
        new_q = self._clone()
        new_q._keywords[keyword] = value
        return new_q

    def filter(self, *criteria):
        new_q = self._clone()
        new_q.criteria += list(criteria)
        return new_q

    def order_by(self, column, order='asc'):
        new_q = self._clone()
        new_q.columns.pop(column,None)
        if order == 'asc':
            new_q.columns[column] = query_number['ORDER_BY']
        elif order == 'desc':
            new_q.columns[column] = query_number['ORDER_BY_DESC']
        else:
            raise ValueError("Ordering must be 'asc' or 'desc'")
        return new_q

    def limit(self, limit):
        new_q = self._clone()
        new_q._limit = limit
        return new_q

    def offset(self, offset):
        new_q = self._clone()
        new_q._offset = offset
        return new_q

    def continue_index(self, continue_index):
        new_q = self._clone()
        new_q._continue_index = continue_index
        return new_q

    def _aggregate(self, func, *args):
        new_q = self._clone()
        # NOTE(wtakase): Override existing column by aggregation.
        for arg in args:
            if isinstance(arg, type) and issubclass(arg, Model):
                for col in arg._columns:
                    self.columns[col] = func
            elif isinstance(arg, Column):
                self.columns[arg] = func
            else:
                raise TypeError("Arguments must be models or columns")
        new_q.columns = self.columns
        return new_q

    def min(self, *args):
        return self._aggregate(query_number['SELECT_MIN'], *args)

    def max(self, *args):
        return self._aggregate(query_number['SELECT_MAX'], *args)

    def sum(self, *args):
        return self._aggregate(query_number['SELECT_SUM'], *args)

    def avg(self, *args):
        return self._aggregate(query_number['SELECT_AVG'], *args)

    def count(self, *args):
        return self._aggregate(query_number['SELECT_COUNT'], *args)

    def _select_message(self):
        dct = OrderedDict([(column.icat_id, value)
                           for (column, value) in six.iteritems(self.columns)])
        return IntegerIntegerMap(dct)

    # todo store criterion for columns and criterion for keywords in seaparate
    # lists
    def _conds_message(self):
        dct = _OrderedMultiMapping([
            (criterion.query_key.icat_id, criterion.op + ' ' + criterion.value)
            for criterion in self.criteria
            if isinstance(criterion.query_key, Column)
        ])
        return IntegerStringMap(dct)

    def _kw_message(self):
        dct = dict([
            (criterion.query_key.icat_key,
             criterion.op + ' ' + criterion.value)
            for criterion in self.criteria
            if isinstance(criterion.query_key, Keyword)
        ])
        for key in self._keywords:
            dct[ key ] = self._keywords[key]
        return StringStringMap(dct)

    def _message(self):
        max_rows = 500 if self._limit == -1 else self._limit
        args = {
            'maxRows': max_rows,
            'continueInx': self._continue_index,
            'partialStartIndex': self._offset,
            'options': 0,
            'KeyValPair_PI': self._kw_message(),
            'InxIvalPair_PI': self._select_message(),
            'InxValPair_PI': self._conds_message()
        }
        return GenQueryRequest(**args)

    def execute(self):
        with self.sess.pool.get_connection() as conn:

            message_body = self._message()
            message = iRODSMessage(
                'RODS_API_REQ', msg=message_body, int_info=api_number['GEN_QUERY_AN'])

            conn.send(message)
            try:
                result_message = conn.recv()
                results = result_message.get_main_message(GenQueryResponse)
                result_set = ResultSet(results)
            except CAT_NO_ROWS_FOUND:
                result_set = ResultSet(
                    empty_gen_query_out(list(self.columns.keys())))
        return result_set

    def close(self):
        '''Closes an open query on the server side.
        self._continue_index must be set to a valid value (returned by a previous query API call).
        '''
        self.limit(0).execute()

    def all(self):
        result_set = self.execute()
        if result_set.continue_index > 0:
            self.continue_index(result_set.continue_index).close()
        return result_set

    def get_batches(self):
        result_set = self.execute()

        try:
            yield result_set

            while result_set.continue_index > 0:
                try:
                    result_set = self.continue_index(
                        result_set.continue_index).execute()
                    yield result_set
                except CAT_NO_ROWS_FOUND:
                    break
        except GeneratorExit:
            if result_set.continue_index > 0:
                self.continue_index(result_set.continue_index).close()

    def get_results(self):
        for result_set in self.get_batches():
            for result in result_set:
                yield result

    def __iter__(self):
        return self.get_results()

    def one(self):
        results = self.execute()
        if results.continue_index > 0:
            self.continue_index(results.continue_index).close()
        if not len(results):
            raise NoResultFound()
        if len(results) > 1:
            raise MultipleResultsFound()
        return results[0]

    def first(self):
        query = self.limit(1)
        results = query.execute()
        if results.continue_index > 0:
            query.continue_index(results.continue_index).close()
        if not len(results):
            return None
        else:
            return results[0]

#     def __getitem__(self, val):
#         pass


class SpecificQuery(object):

    def __init__(self, sess, sql=None, alias=None, columns=None, args=None):
        if not sql and not alias:
            raise ValueError('A query or alias must be provided')

        self.session = sess
        self._sql = sql
        self._alias = alias
        self._continue_index = 0
        self._columns = columns
        self._args = args or []


    def register(self):
        if not self._sql:
            raise ValueError('Empty query')

        message_body = GeneralAdminRequest(
            "add",
            "specificQuery",
            self._sql,
            self._alias
        )
        request = iRODSMessage("RODS_API_REQ", msg=message_body,
                               int_info=api_number['GENERAL_ADMIN_AN'])

        with self.session.pool.get_connection() as conn:
            conn.send(request)
            response = conn.recv()
        return response


    def remove(self):
        target =  self._alias or self._sql

        message_body = GeneralAdminRequest(
            "rm",
            "specificQuery",
            target
        )
        request = iRODSMessage("RODS_API_REQ", msg=message_body,
                               int_info=api_number['GENERAL_ADMIN_AN'])

        with self.session.pool.get_connection() as conn:
            conn.send(request)
            response = conn.recv()
        return response


    def execute(self, limit=MAX_SQL_ROWS, offset=0, options=0, conditions=None):
        target =  self._alias or self._sql

        if conditions is None:
            conditions = StringStringMap({})

        sql_args = {}
        for i, arg in enumerate(self._args[:10], start=1):
            sql_args['arg{}'.format(i)] = arg

        message_body = SpecificQueryRequest(sql=target,
                                            maxRows=limit,
                                            continueInx=self._continue_index,
                                            rowOffset=offset,
                                            options=0,
                                            KeyValPair_PI=conditions,
                                            **sql_args)

        request = iRODSMessage("RODS_API_REQ", msg=message_body, int_info=api_number['SPECIFIC_QUERY_AN'])

        with self.session.pool.get_connection() as conn:
            conn.send(request)
            response = conn.recv()

        results = response.get_main_message(GenQueryResponse)
        return SpecificQueryResultSet(results, self._columns)


    def __iter__(self):
        return self.get_results()


    def get_batches(self):
        result_set = self.execute()
        yield result_set

        while result_set.continue_index > 0:
            self._continue_index = result_set.continue_index
            try:
                result_set = self.execute()
                yield result_set
            except CAT_NO_ROWS_FOUND:
                break


    def get_results(self):
        for result_set in self.get_batches():
            for result in result_set:
                yield result




from __future__ import absolute_import
from irods.models import User, UserGroup, UserAuth
from irods.meta import iRODSMetaCollection
from irods.exception import NoResultFound

_Not_Defined = ()

class iRODSUser(object):

    def __init__(self, manager, result=None):
        self.manager = manager
        if result:
            self.id = result[User.id]
            self.name = result[User.name]
            self.type = result[User.type]
            self.zone = result[User.zone]
            self._comment = result.get(User.comment, _Not_Defined)  # these not needed in results for object ident,
            self._info = result.get(User.info, _Not_Defined)        # so we fetch lazily via a property
        self._meta = None

    @property
    def comment(self):
        if self._comment == _Not_Defined:
            query = self.manager.sess.query(User.id,User.comment).filter(User.id == self.id)
            self._comment = query.one()[User.comment]
        return self._comment

    @property
    def info(self):
        if self._info == _Not_Defined:
            query = self.manager.sess.query(User.id,User.info).filter(User.id == self.id)
            self._info = query.one()[User.info]
        return self._info

    @property
    def dn(self):
        query = self.manager.sess.query(UserAuth.user_dn).filter(UserAuth.user_id == self.id)
        return [res[UserAuth.user_dn] for res in query]

    @property
    def metadata(self):
        if not self._meta:
            self._meta = iRODSMetaCollection(
                 self.manager.sess.metadata, User, self.name)
        return self._meta

    def modify(self, *args, **kwargs):
        self.manager.modify(self.name, *args, **kwargs)

    def __repr__(self):
        return "<iRODSUser {id} {name} {type} {zone}>".format(**vars(self))

    def remove(self):
        self.manager.remove(self.name, self.zone)

    def temp_password(self):
        return self.manager.temp_password_for_user(self.name)


class iRODSUserGroup(object):

    def __init__(self, manager, result=None):
        self.manager = manager
        if result:
            self.id = result[UserGroup.id]
            self.name = result[UserGroup.name]
        self._meta = None

    def __repr__(self):
        return "<iRODSUserGroup {id} {name}>".format(**vars(self))

    def remove(self):
        self.manager.remove(self.name)

    @property
    def members(self):
        return self.manager.getmembers(self.name)

    @property
    def metadata(self):
        if not self._meta:
            self._meta = iRODSMetaCollection(
                 self.manager.sess.metadata, User, self.name)
        return self._meta

    def addmember(self, user_name, user_zone=""):
        self.manager.addmember(self.name, user_name, user_zone)

    def removemember(self, user_name, user_zone=""):
        self.manager.removemember(self.name, user_name, user_zone)

    def hasmember(self, user_name):
        member_names = [user.name for user in self.members]
        return user_name in member_names



#!/usr/bin/env python
from __future__ import print_function

import os
import ssl
import time
import sys
import logging
import contextlib
import concurrent.futures
import threading
import multiprocessing
import six

from irods.data_object import iRODSDataObject
from irods.exception import DataObjectDoesNotExist
import irods.keywords as kw
from six.moves.queue import Queue,Full,Empty


logger = logging.getLogger( __name__ )
_nullh  = logging.NullHandler()
logger.addHandler( _nullh )


MINIMUM_SERVER_VERSION = (4,2,9)


try:
    from threading import Barrier   # Use 'Barrier' class if included (as in Python >= 3.2) ...
except ImportError:                 # ... but otherwise, use this ad hoc:
    class Barrier(object):
        def __init__(self, n):
            """Initialize a Barrier to wait on n threads."""
            self.n = n
            self.count = 0
            self.mutex = threading.Semaphore(1)
            self.barrier = threading.Semaphore(0)
        def wait(self):
            """Per-thread wait function.

            As in Python3.2 threading, returns 0 <= wait_serial_int < n
            """
            self.mutex.acquire()
            self.count += 1
            count = self.count
            self.mutex.release()
            if count == self.n: self.barrier.release()
            self.barrier.acquire()
            self.barrier.release()
            return count - 1

@contextlib.contextmanager
def enableLogging(handlerType,args,level_ = logging.INFO):
    """Context manager for temporarily enabling a logger. For debug or test.

    Usage Example -
    with irods.parallel.enableLogging(logging.FileHandler,('/tmp/logfile.txt',)):
        # parallel put/get code here
    """
    h = None
    saveLevel = logger.level
    try:
        logger.setLevel(level_)
        h = handlerType(*args)
        h.setLevel( level_ )
        logger.addHandler(h)
        yield
    finally:
        logger.setLevel(saveLevel)
        if h in logger.handlers:
            logger.removeHandler(h)


RECOMMENDED_NUM_THREADS_PER_TRANSFER = 3

verboseConnection = False

class BadCallbackTarget(TypeError): pass

class AsyncNotify (object):

    """A type returned when the PUT or GET operation passed includes NONBLOCKING.
       If enabled, the callback function (or callable object) will be triggered
       when all parts of the parallel transfer are complete.  It should accept
       exactly one argument, the irods.parallel.AsyncNotify instance that
       is calling it.
    """

    def set_transfer_done_callback( self, callback ):
        if callback is not None:
            if not callable(callback):
                raise BadCallbackTarget( '"callback" must be a callable accepting at least 1 argument' )
        self.done_callback = callback

    def __init__(self, futuresList, callback = None, progress_Queue = None, total = None, keep_ = ()):
        """AsyncNotify initialization (used internally to the io.parallel library).
           The casual user will only be concerned with the callback parameter, called when all threads
           of the parallel PUT or GET have been terminated and the data object closed.
        """
        self._futures = set(futuresList)
        self._futures_done = dict()
        self.keep = dict(keep_)
        self._lock = threading.Lock()
        self.set_transfer_done_callback (callback)
        self.__done = False
        if self._futures:
            for future in self._futures: future.add_done_callback( self )
        else:
            self.__invoke_done_callback()

        self.progress = [0, 0]
        if (progress_Queue) and (total is not None):
            self.progress[1] = total
            def _progress(Q,this):  # - thread to update progress indicator
                while this.progress[0] < this.progress[1]:
                    i = None
                    try:
                        i = Q.get(timeout=0.1)
                    except Empty:
                        pass
                    if i is not None:
                        if isinstance(i,six.integer_types) and i >= 0: this.progress[0] += i
                        else: break
            self._progress_fn = _progress
            self._progress_thread = threading.Thread( target = self._progress_fn, args = (progress_Queue, self))
            self._progress_thread.start()

    @staticmethod
    def asciiBar( lst, memo = [1] ):
        memo[0] += 1
        spinner = "|/-\\"[memo[0]%4]
        percent = "%5.1f%%"%(lst[0]*100.0/lst[1])
        mbytes = "%9.1f MB / %9.1f MB"%(lst[0]/1e6,lst[1]/1e6)
        if lst[1] != 0:
            s = "  {spinner} {percent} [ {mbytes} ] "
        else:
            s = "  {spinner} "
        return s.format(**locals())

    def wait_until_transfer_done (self, timeout=float('inf'), progressBar = False):
        carriageReturn = '\r'
        begin = t = time.time()
        end = begin + timeout
        while not self.__done:
            time.sleep(min(0.1, max(0.0, end - t)))
            t = time.time()
            if t >= end: break
            if progressBar:
                print ('  ' + self.asciiBar( self.progress ) + carriageReturn, end='', file=sys.stderr)
                sys.stderr.flush()
        return self.__done

    def __call__(self,future): # Our instance is called by each future (individual file part) when done.
                               # When all futures are done, we invoke the configured callback.
        with self._lock:
            self._futures_done[future] = future.result()
            if len(self._futures) == len(self._futures_done): self.__invoke_done_callback()

    def __invoke_done_callback(self):
        try:
            if callable(self.done_callback): self.done_callback(self)
        finally:
            self.keep.pop('mgr',None)
            self.__done = True
        self.set_transfer_done_callback(None)

    @property
    def futures(self): return list(self._futures)

    @property
    def futures_done(self): return dict(self._futures_done)


class Oper(object):
    """A custom enum-type class with utility methods.  """

    GET = 0
    PUT = 1
    NONBLOCKING = 2

    def __int__(self):
        """Return the stored flags as an integer bitmask. """
        return self._opr

    def __init__(self, rhs):
        """Initialize with a bit mask of flags ie. whether Operation PUT or GET, 
        and whether NONBLOCKING."""
        self._opr = int(rhs)

    def isPut(self): return 0 != (self._opr & self.PUT)
    def isGet(self): return not self.isPut()
    def isNonBlocking(self): return 0 != (self._opr & self.NONBLOCKING)

    def data_object_mode(self, initial_open = False):
        if self.isPut():
            return 'w' if initial_open else 'a'
        else:
            return 'r'

    def disk_file_mode(self, initial_open = False, binary = True):
        if self.isPut():
            mode = 'r'
        else:
            mode = 'w' if initial_open else 'r+'
        return ((mode + 'b') if binary else mode)


def _io_send_bytes_progress (queueObject, item):
    try:
        queueObject.put(item)
        return True
    except Full:
        return False

COPY_BUF_SIZE = (1024 ** 2) * 4

def _copy_part( src, dst, length, queueObject, debug_info, mgr):

    bytecount = 0
    accum = 0
    while True and bytecount < length:
        buf = src.read(min(COPY_BUF_SIZE, length - bytecount))
        buf_len = len(buf)
        if 0 == buf_len: break
        dst.write(buf)
        bytecount += buf_len
        accum += buf_len
        if queueObject and accum and _io_send_bytes_progress(queueObject,accum): accum = 0
        if verboseConnection:
            print ("("+debug_info+")",end='',file=sys.stderr)
            sys.stderr.flush()

    # In a put or get, exactly one of (src,dst) is a file. Find which and close that one first.
    (file_,obj_) = (src,dst) if dst in mgr else (dst,src)
    file_.close()
    mgr.remove_io( obj_ ) # 1. closes obj if it is not the mgr's initial descriptor
                          # 2. blocks at barrier until all transfer threads are done copying
                          # 3. closes with finalize if obj is mgr's initial descriptor
    return bytecount


class _Multipart_close_manager:

    def __init__(self, initial_io_, exit_barrier_):
        self.exit_barrier = exit_barrier_
        self.initial_io = initial_io_
        self.__lock = threading.Lock()
        self.aux = []

    def __contains__(self,Io):
        with self.__lock:
            return Io is self.initial_io or \
                   Io in self.aux

    # `add_io' - add an i/o object to be managed
    # note: `remove_io' should only be called for managed i/o objects

    def add_io(self,Io):
        with self.__lock:
            if Io is not self.initial_io:
                self.aux.append(Io)

    # `remove_io' is for closing a channel of parallel i/o and allowing the
    # data object to flush write operations (if any) in a timely fashion.  It also
    # synchronizes all of the parallel threads just before exit, so that we know
    # exactly when to perform a finalizing close on the data object

    def remove_io(self,Io):
        is_initial = True
        with self.__lock:
            if Io is not self.initial_io:
                Io.close()
                self.aux.remove(Io)
                is_initial = False
        self.exit_barrier.wait()
        if is_initial: self.finalize()

    def finalize(self):
        self.initial_io.close()


def _io_part (objHandle, range_, file_, opr_, mgr_, thread_debug_id = '', queueObject = None ):
    if 0 == len(range_): return 0
    Operation = Oper(opr_)
    (offset,length) = (range_[0], len(range_))
    objHandle.seek(offset)
    file_.seek(offset)
    if thread_debug_id == '':
        thread_debug_id = str(threading.currentThread().ident)
    return ( _copy_part (file_, objHandle, length, queueObject, thread_debug_id, mgr_) if Operation.isPut()
        else _copy_part (objHandle, file_, length, queueObject, thread_debug_id, mgr_) )


def _io_multipart_threaded(operation_ , dataObj_and_IO, replica_token, hier_str, session, fname,
                           total_size, num_threads = 0, **extra_options):
    """Called by _io_main.
    Carve up (0,total_size) range into `num_threads` parts and initiate a transfer thread for each one."""
    (D, Io) = dataObj_and_IO
    Operation = Oper( operation_ )

    if num_threads < 1:
        num_threads = RECOMMENDED_NUM_THREADS_PER_TRANSFER
    num_threads = max(1, min(multiprocessing.cpu_count(), num_threads))

    P = 1 + (total_size // num_threads)
    logger.info("num_threads = %s ; (P)artitionSize = %s", num_threads, P)
    ranges = [six.moves.range(i*P,min(i*P+P,total_size)) for i in range(num_threads) if i*P < total_size]

    _queueLength = extra_options.get('_queueLength',0)
    if _queueLength > 0:
        queueObject = Queue(_queueLength)
    else:
        queueObject = None

    futures = []
    executor = concurrent.futures.ThreadPoolExecutor(max_workers = num_threads)
    num_threads = min(num_threads, len(ranges))
    mgr = _Multipart_close_manager(Io, Barrier(num_threads))
    counter = 1
    gen_file_handle = lambda: open(fname, Operation.disk_file_mode(initial_open = (counter == 1)))
    File = gen_file_handle()
    for r in ranges:
        if Io is None:
            Io = session.data_objects.open( D.path, Operation.data_object_mode(initial_open = False),
                                            create = False, finalize_on_close = False,
                                            **{kw.RESC_HIER_STR_KW: hier_str, kw.REPLICA_TOKEN_KW: replica_token} )
        mgr.add_io( Io )
        if File is None: File = gen_file_handle()
        futures.append(executor.submit( _io_part, Io, r, File, Operation, mgr, str(counter), queueObject))
        counter += 1
        Io = File = None

    if Operation.isNonBlocking():
        if _queueLength:
            return futures, queueObject, mgr
        else:
            return futures
    else:
        bytecounts = [ f.result() for f in futures ]
        return sum(bytecounts), total_size


# _io_main
#    * Entry point for parallel transfers (multithreaded PUT and GET operations) 
#    * determine replica information
#    * call multithread manager

def io_main( session, Data, opr_, fname, R='', **kwopt):

    Operation = Oper(opr_)
    d_path = None
    Io = None
    if isinstance(Data,tuple):
        (Data, Io) = Data[:2]
    if isinstance (Data, six.string_types):
        d_path = Data
        try:
            Data = session.data_objects.get( Data )
            d_path = Data.path
        except DataObjectDoesNotExist:
            if Operation.isGet(): raise

    R_via_libcall = kwopt.pop( 'target_resource_name', '')
    if R_via_libcall:
        R = R_via_libcall

    resc_options = {}
    if Operation.isPut():
        if R:
            resc_options [kw.RESC_NAME_KW] = R
            resc_options [kw.DEST_RESC_NAME_KW] = R

    if (not Io):
        (Io, rawfile) = session.data_objects.open_with_FileRaw( (d_path or Data.path),
                                                                Operation.data_object_mode(initial_open = True),
                                                                finalize_on_close = True, **resc_options )
    else:
        rawfile = Io.raw

    # data object should now exist
    if not isinstance(Data,iRODSDataObject):
        Data = session.data_objects.get(d_path)

    if Operation.isGet():
        total_bytes = Io.seek(0,os.SEEK_END)
        Io.seek(0,os.SEEK_SET)
    else:
        with open(fname, 'rb') as f:
            f.seek(0,os.SEEK_END)
            total_bytes = f.tell()

    (replica_token , resc_hier) = rawfile.replica_access_info()

    num_threads = kwopt.pop( 'num_threads', None)

    if num_threads is None: num_threads = int(kwopt.get('N','0'))

    queueLength = kwopt.get('queueLength',0)
    retval = _io_multipart_threaded (Operation, (Data, Io), replica_token, resc_hier, session, fname, total_bytes,
                                     num_threads = num_threads,
                                     _queueLength = queueLength)

    # SessionObject.data_objects.parallel_{put,get} will return:
    #   - immediately with an AsyncNotify instance, if Oper.NONBLOCKING flag is used.
    #   - upon completion with a boolean completion status, otherwise.

    if Operation.isNonBlocking():

        if queueLength > 0:
            (futures, chunk_notify_queue, mgr) = retval
        else:
            futures = retval
            chunk_notify_queue = total_bytes = None

        return AsyncNotify( futures,                              # individual futures, one per transfer thread
                            progress_Queue = chunk_notify_queue,  # for notifying the progress indicator thread
                            total = total_bytes,                  # total number of bytes for parallel transfer
                            keep_ = {'mgr': mgr}  )   # an open raw i/o object needing to be persisted, if any
    else:
        (_bytes_transferred, _bytes_total) = retval
        return (_bytes_transferred == _bytes_total)

if __name__ == '__main__':

    import getopt
    import atexit
    from irods.session import iRODSSession

    def setupLoggingWithDateTimeHeader(name,level = logging.DEBUG):
        if _nullh in logger.handlers:
            logger.removeHandler(_nullh)
            if name:
                handler = logging.FileHandler(name)
            else:
                handler = logging.StreamHandler()
            handler.setFormatter(logging.Formatter('%(asctime)-15s - %(message)s'))
        logger.addHandler(handler)
        logger.setLevel( level )

    try:
        env_file = os.environ['IRODS_ENVIRONMENT_FILE']
    except KeyError:
        env_file = os.path.expanduser('~/.irods/irods_environment.json')
    ssl_context = ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH, cafile=None, capath=None, cadata=None)
    ssl_settings = {'ssl_context': ssl_context}
    sess = iRODSSession(irods_env_file=env_file, **ssl_settings)
    atexit.register(lambda : sess.cleanup())

    opt,arg = getopt.getopt( sys.argv[1:], 'vL:l:aR:N:')

    opts = dict(opt)

    logFilename = opts.pop('-L',None)  # '' for console, non-empty for filesystem destination
    logLevel = (logging.INFO if logFilename is None else logging.DEBUG)
    logFilename = logFilename or opts.pop('-l',None)

    if logFilename is not None:
        setupLoggingWithDateTimeHeader(logFilename, logLevel)

    verboseConnection = (opts.pop('-v',None) is not None)

    async_xfer = opts.pop('-a',None)

    kwarg = { k.lstrip('-'):v for k,v in opts.items() }

    arg[1] = Oper.PUT if arg[1].lower() in ('w','put','a') \
                      else Oper.GET
    if async_xfer is not None:
        arg[1] |= Oper.NONBLOCKING

    ret = io_main(sess, *arg, **kwarg) # arg[0] = data object or path
                                       # arg[1] = operation: or'd flags : [PUT|GET] NONBLOCKING
                                       # arg[2] = file path on local filesystem
                                       # kwarg['queueLength'] sets progress-queue length (0 if no progress indication needed)
                                       # kwarg options 'N' (num threads) and 'R' (target resource name) are via command-line
                                       # kwarg['num_threads'] (overrides 'N' when called as a library)
                                       # kwarg['target_resource_name'] (overrides 'R' when called as a library)
    if isinstance( ret, AsyncNotify ):
        print('waiting on completion...',file=sys.stderr)
        ret.set_transfer_done_callback(lambda r: print('Async transfer done for:',r,file=sys.stderr))
        done = ret.wait_until_transfer_done (timeout=10.0)  # - or do other useful work here
        if done:
            bytes_transferred = sum(ret.futures_done.values())
            print ('Asynch transfer complete. Total bytes transferred:', bytes_transferred,file=sys.stderr)
        else:
            print ('Asynch transfer was not completed before timeout expired.',file=sys.stderr)
    else:
        print('Synchronous transfer {}'.format('succeeded' if ret else 'failed'),file=sys.stderr)

# Note : This module requires concurrent.futures, included in Python3.x.
#        On Python2.7, this dependency must be installed using 'pip install futures'.
# Demonstration :
#
# $ dd if=/dev/urandom bs=1k count=150000 of=$HOME/puttest
# $ time python -m irods.parallel -R demoResc -N 3 `ipwd`/test.dat put $HOME/puttest  # add -v,-a for verbose, asynch
# $ time python -m irods.parallel -R demoResc -N 3 `ipwd`/test.dat get $HOME/gettest  # add -v,-a for verbose, asynch
# $ diff puttest gettest



#! /usr/bin/env python
from __future__ import absolute_import
import datetime
import os
import re
import sys
import time
import json
import unittest
import irods.test.helpers as helpers

#  Regular expression to match common synonyms for localhost.
#

LOCALHOST_REGEX = re.compile(r"""^(127(\.\d+){1,3}|[0:]+1|(.*-)?localhost(\.\w+)?)$""",re.IGNORECASE)
USE_ONLY_LOCALHOST = False


class TestPool(unittest.TestCase):

    config_extension = ".json"
    test_extension = ""
    preferred_parameters = {}

    @classmethod
    def setUpClass(cls):              # generate test env files using connect data from ~/.irods environment
        if USE_ONLY_LOCALHOST: return
        Nonlocal_Ext = ".test"
        with helpers.make_session() as session:
            cls.preferred_parameters = { 'irods_host':session.host,
                                         'irods_port':session.port,
                                         'irods_user_name':session.username,
                                         'irods_zone_name':session.zone }
            test_configs_dir = os.path.join(irods_test_path(),"test-data")
            for config in [os.path.join(test_configs_dir,f) for f in os.listdir(test_configs_dir)
                           if f.endswith(cls.config_extension)]:
                with open(config,"r") as in_, open(config + Nonlocal_Ext,"w") as out_:
                    cf = json.load(in_)
                    cf.update(cls.preferred_parameters)
                    json.dump(cf, out_,indent=4)
            cls.test_extension = Nonlocal_Ext


    def setUp(self):
        self.sess = helpers.make_session(
            irods_env_file=os.path.join(irods_test_path(),"test-data","irods_environment.json" + self.test_extension))
        if USE_ONLY_LOCALHOST and not LOCALHOST_REGEX.match (self.sess.host):
            self.skipTest('for non-local server')

    def tearDown(self):
        '''Close connections
        '''
        self.sess.cleanup()

    def test_release_connection(self):
        with self.sess.pool.get_connection():
            self.assertEqual(1, len(self.sess.pool.active))
            self.assertEqual(0, len(self.sess.pool.idle))

        self.assertEqual(0, len(self.sess.pool.active))
        self.assertEqual(1, len(self.sess.pool.idle))

    def test_destroy_active(self):
        with self.sess.pool.get_connection() as conn:
            self.assertEqual(1, len(self.sess.pool.active))
            self.assertEqual(0, len(self.sess.pool.idle))
            conn.release(destroy=True)

        self.assertEqual(0, len(self.sess.pool.active))
        self.assertEqual(0, len(self.sess.pool.idle))

    def test_destroy_idle(self):
        with self.sess.pool.get_connection():
            self.assertEqual(1, len(self.sess.pool.active))
            self.assertEqual(0, len(self.sess.pool.idle))

        # cleanup all connections
        self.sess.cleanup()
        self.assertEqual(0, len(self.sess.pool.active))
        self.assertEqual(0, len(self.sess.pool.idle))

    def test_release_disconnected(self):
        with self.sess.pool.get_connection() as conn:
            self.assertEqual(1, len(self.sess.pool.active))
            self.assertEqual(0, len(self.sess.pool.idle))
            conn.disconnect()

        # even though disconnected, gets put into idle
        self.assertEqual(0, len(self.sess.pool.active))
        self.assertEqual(1, len(self.sess.pool.idle))

        # should remove all connections
        self.sess.cleanup()
        self.assertEqual(0, len(self.sess.pool.active))
        self.assertEqual(0, len(self.sess.pool.idle))

    def test_connection_create_time(self):
        # Get a connection and record its object ID and create_time
        # Release the connection (goes from active to idle queue)
        # Again, get a connection. Should get the same connection back.
        # I.e., the object IDs should match. However, the new connection
        # should have a more recent 'last_used_time'
        conn_obj_id_1 = None
        conn_obj_id_2 = None
        create_time_1 = None
        create_time_2 = None
        last_used_time_1 = None
        last_used_time_2 = None

        with self.sess.pool.get_connection() as conn:
            conn_obj_id_1 = id(conn)
            curr_time = datetime.datetime.now()
            create_time_1 = conn.create_time
            last_used_time_1 = conn.last_used_time
            self.assertTrue(curr_time >= create_time_1)
            self.assertTrue(curr_time >= last_used_time_1)
            self.assertEqual(1, len(self.sess.pool.active))
            self.assertEqual(0, len(self.sess.pool.idle))

            self.sess.pool.release_connection(conn)
            self.assertEqual(0, len(self.sess.pool.active))
            self.assertEqual(1, len(self.sess.pool.idle))

        with self.sess.pool.get_connection() as conn:
            conn_obj_id_2 = id(conn)
            curr_time = datetime.datetime.now()
            create_time_2 = conn.create_time
            last_used_time_2 = conn.last_used_time
            self.assertEqual(conn_obj_id_1, conn_obj_id_2)
            self.assertTrue(curr_time >= create_time_2)
            self.assertTrue(curr_time >= last_used_time_2)
            self.assertTrue(last_used_time_2 >= last_used_time_1)
            self.assertEqual(1, len(self.sess.pool.active))
            self.assertEqual(0, len(self.sess.pool.idle))

            self.sess.pool.release_connection(conn)
            self.assertEqual(0, len(self.sess.pool.active))
            self.assertEqual(1, len(self.sess.pool.idle))

            self.sess.pool.release_connection(conn, True)
            self.assertEqual(0, len(self.sess.pool.active))
            self.assertEqual(0, len(self.sess.pool.idle))

    def test_refresh_connection(self):
        # Set 'irods_connection_refresh_time' to '3' (in seconds) in
        # ~/.irods/irods_environment.json file. This means any connection
        # that was created more than 3 seconds ago will be dropped and
        # a new connection is created/returned. This is to avoid
        # issue with idle connections that are dropped.
        conn_obj_id_1 = None
        conn_obj_id_2 = None
        create_time_1 = None
        create_time_2 = None
        last_used_time_1 = None
        last_used_time_2 = None

        with self.sess.pool.get_connection() as conn:
            conn_obj_id_1 = id(conn)
            curr_time = datetime.datetime.now()
            create_time_1 = conn.create_time
            last_used_time_1 = conn.last_used_time
            self.assertTrue(curr_time >= create_time_1)
            self.assertTrue(curr_time >= last_used_time_1)
            self.assertEqual(1, len(self.sess.pool.active))
            self.assertEqual(0, len(self.sess.pool.idle))

            self.sess.pool.release_connection(conn)
            self.assertEqual(0, len(self.sess.pool.active))
            self.assertEqual(1, len(self.sess.pool.idle))

        # Wait more than 'irods_connection_refresh_time' seconds,
        # which is set to 3. Connection object should have a different
        # object ID (as a new connection is created)
        time.sleep(5)

        with self.sess.pool.get_connection() as conn:
            conn_obj_id_2 = id(conn)
            curr_time = datetime.datetime.now()
            create_time_2 = conn.create_time
            last_used_time_2 = conn.last_used_time
            self.assertTrue(curr_time >= create_time_2)
            self.assertTrue(curr_time >= last_used_time_2)
            self.assertNotEqual(conn_obj_id_1, conn_obj_id_2)
            self.assertTrue(create_time_2 > create_time_1)
            self.assertEqual(1, len(self.sess.pool.active))
            self.assertEqual(0, len(self.sess.pool.idle))

            self.sess.pool.release_connection(conn, True)
            self.assertEqual(0, len(self.sess.pool.active))
            self.assertEqual(0, len(self.sess.pool.idle))

    def test_no_refresh_connection(self):
        # Set 'irods_connection_refresh_time' to '3' (in seconds) in
        # ~/.irods/irods_environment.json file. This means any connection
        # created more than 3 seconds ago will be dropped and
        # a new connection is created/returned. This is to avoid
        # issue with idle connections that are dropped.
        conn_obj_id_1 = None
        conn_obj_id_2 = None
        create_time_1 = None
        create_time_2 = None
        last_used_time_1 = None
        last_used_time_2 = None

        with self.sess.pool.get_connection() as conn:
            conn_obj_id_1 = id(conn)
            curr_time = datetime.datetime.now()
            create_time_1 = conn.create_time
            last_used_time_1 = conn.last_used_time
            self.assertTrue(curr_time >= create_time_1)
            self.assertTrue(curr_time >= last_used_time_1)
            self.assertEqual(1, len(self.sess.pool.active))
            self.assertEqual(0, len(self.sess.pool.idle))

            self.sess.pool.release_connection(conn)
            self.assertEqual(0, len(self.sess.pool.active))
            self.assertEqual(1, len(self.sess.pool.idle))

        # Wait less than 'irods_connection_refresh_time' seconds,
        # which is set to 3. Connection object should have the same
        # object ID (as idle time is less than 'irods_connection_refresh_time')
        time.sleep(1)

        with self.sess.pool.get_connection() as conn:
            conn_obj_id_2 = id(conn)
            curr_time = datetime.datetime.now()
            create_time_2 = conn.create_time
            last_used_time_2 = conn.last_used_time
            self.assertTrue(curr_time >= create_time_2)
            self.assertTrue(curr_time >= last_used_time_2)
            self.assertEqual(conn_obj_id_1, conn_obj_id_2)
            self.assertTrue(create_time_2 >= create_time_1)
            self.assertEqual(1, len(self.sess.pool.active))
            self.assertEqual(0, len(self.sess.pool.idle))

            self.sess.pool.release_connection(conn, True)
            self.assertEqual(0, len(self.sess.pool.active))
            self.assertEqual(0, len(self.sess.pool.idle))


    def test_get_connection_refresh_time_no_env_file_input_param(self):
        connection_refresh_time = self.sess.get_connection_refresh_time(first_name="Magic", last_name="Johnson")
        self.assertEqual(connection_refresh_time, -1)

    def test_get_connection_refresh_time_none_existant_env_file(self):
        connection_refresh_time = self.sess.get_connection_refresh_time(
            irods_env_file=os.path.join(irods_test_path(),"test-data","irods_environment_non_existant.json" + self.test_extension))
        self.assertEqual(connection_refresh_time, -1)

    def test_get_connection_refresh_time_no_connection_refresh_field(self):
        connection_refresh_time = self.sess.get_connection_refresh_time(
            irods_env_file=os.path.join(irods_test_path(),"test-data","irods_environment_no_refresh_field.json" + self.test_extension))
        self.assertEqual(connection_refresh_time, -1)

    def test_get_connection_refresh_time_negative_connection_refresh_field(self):
        connection_refresh_time = self.sess.get_connection_refresh_time(
            irods_env_file=os.path.join(irods_test_path(),"test-data","irods_environment_negative_refresh_field.json" + self.test_extension))
        self.assertEqual(connection_refresh_time, -1)

    def test_get_connection_refresh_time(self):
        default_path = os.path.join (irods_test_path(),"test-data","irods_environment.json" + self.test_extension)
        connection_refresh_time = self.sess.get_connection_refresh_time(irods_env_file=default_path)
        self.assertEqual(connection_refresh_time, 3)


def irods_test_path():
    return os.path.dirname(__file__)


if __name__ == '__main__':
    # let the tests find the parent irods lib
    sys.path.insert(0, os.path.abspath('../..'))
    unittest.main()



#! /usr/bin/env python
from __future__ import absolute_import
import os
import sys
import socket
import shutil
import unittest
from irods.meta import iRODSMetaCollection
from irods.exception import CollectionDoesNotExist
from irods.models import Collection, DataObject
import irods.test.helpers as helpers
import irods.keywords as kw
from six.moves import range


class TestCollection(unittest.TestCase):

    def setUp(self):
        self.sess = helpers.make_session()
        self.test_coll_path = '/{}/home/{}/test_dir'.format(self.sess.zone, self.sess.username)

        self.test_coll = self.sess.collections.create(self.test_coll_path)


    def tearDown(self):
        """ Delete the test collection after each test """
        self.test_coll.remove(recurse=True, force=True)
        self.sess.cleanup()


    def test_get_collection(self):
        # path = "/tempZone/home/rods"
        coll = self.sess.collections.get(self.test_coll_path)
        self.assertEqual(self.test_coll_path, coll.path)

    def test_irods_collection_information(self):
        coll = self.sess.collections.get(self.test_coll_path)
        self.assertIsNotNone(coll.create_time)
        self.assertIsNotNone(coll.modify_time)
        self.assertFalse(coll.inheritance)
        self.assertIsNotNone(coll.owner_name)
        self.assertIsNotNone(coll.owner_zone)

    def test_append_to_collection(self):
        """ Append a new file to the collection"""
        pass


    def test_remove_from_collection(self):
        """ Delete a file from a collection """
        pass


    def test_update_in_collection(self):
        """ Modify a file in a collection """
        pass


    def test_create_recursive_collection(self):
        # make path with recursion
        root_coll_path = self.test_coll_path + "/recursive/collection/test"
        self.sess.collections.create(root_coll_path, recurse=True)

        #confirm col create
        coll = self.sess.collections.get(root_coll_path)
        self.assertEqual(root_coll_path, coll.path)

        # delete test collection
        coll.remove(force=True)

        # confirm delete
        with self.assertRaises(CollectionDoesNotExist):
            self.sess.collections.get(root_coll_path)

    def test_remove_deep_collection(self):
        # depth = 100
        depth = 20  # placeholder
        root_coll_path = self.test_coll_path + "/deep_collection"

        # make test collection
        helpers.make_deep_collection(
            self.sess, root_coll_path, depth=depth, objects_per_level=1, object_content=None)

        # delete test collection
        self.sess.collections.remove(root_coll_path, recurse=True, force=True)

        # confirm delete
        with self.assertRaises(CollectionDoesNotExist):
            self.sess.collections.get(root_coll_path)


    def test_rename_collection(self):
        # test args
        args = {'collection': self.test_coll_path,
                'old_name': 'foo',
                'new_name': 'bar'}

        # make collection
        path = "{collection}/{old_name}".format(**args)
        coll = helpers.make_collection(self.sess, path)

        # get collection id
        saved_id = coll.id

        # rename coll
        new_path = "{collection}/{new_name}".format(**args)
        coll.move(new_path)
        # self.sess.collections.move(path, new_path)

        # get updated collection
        coll = self.sess.collections.get(new_path)

        # compare ids
        self.assertEqual(coll.id, saved_id)

        # remove collection
        coll.remove(recurse=True, force=True)


    def test_move_coll_to_coll(self):
        # test args
        args = {'base_collection': self.test_coll_path,
                'collection1': 'foo',
                'collection2': 'bar'}

        # make collection1 and collection2 in base collection
        path1 = "{base_collection}/{collection1}".format(**args)
        coll1 = helpers.make_collection(self.sess, path1)
        path2 = "{base_collection}/{collection2}".format(**args)
        coll2 = helpers.make_collection(self.sess, path2)

        # get collection2's id
        saved_id = coll2.id

        # move collection2 into collection1
        self.sess.collections.move(path2, path1)

        # get updated collection
        path2 = "{base_collection}/{collection1}/{collection2}".format(**args)
        coll2 = self.sess.collections.get(path2)

        # compare ids
        self.assertEqual(coll2.id, saved_id)

        # remove collection
        coll1.remove(recurse=True, force=True)


    def test_repr_coll(self):
        coll_name = self.test_coll.name.encode('utf-8')
        coll_id = self.test_coll.id

        self.assertEqual(
            repr(self.test_coll), "<iRODSCollection {coll_id} {coll_name}>".format(**locals()))


    def test_walk_collection_topdown(self):
        depth = 20

        # files that will be ceated in each subcollection
        filenames = ['foo', 'bar', 'baz']

        # make nested collections
        coll_path = self.test_coll_path
        for d in range(depth):
            # create subcollection with files
            coll_path += '/sub' + str(d)
            helpers.make_collection(self.sess, coll_path, filenames)

        # now walk nested collections
        colls = self.test_coll.walk()
        current_coll_name = self.test_coll.name
        for d in range(depth + 1):
            # get next result
            collection, subcollections, data_objects = next(colls)

            # check collection name
            self.assertEqual(collection.name, current_coll_name)

            # check subcollection name
            if d < depth:
                sub_coll_name = 'sub' + str(d)
                self.assertEqual(sub_coll_name, subcollections[0].name)
            else:
                # last coll has no subcolls
                self.assertListEqual(subcollections, [])

            # check data object names
            for data_object in data_objects:
                self.assertIn(data_object.name, filenames)

            # iterate
            current_coll_name = sub_coll_name

        # that should be it
        with self.assertRaises(StopIteration):
            next(colls)


    def test_walk_collection(self):
        depth = 20

        # files that will be ceated in each subcollection
        filenames = ['foo', 'bar', 'baz']

        # make nested collections
        coll_path = self.test_coll_path
        for d in range(depth):
            # create subcollection with files
            coll_path += '/sub' + str(d)
            helpers.make_collection(self.sess, coll_path, filenames)

        # now walk nested collections
        colls = self.test_coll.walk(topdown=False)
        sub_coll_name = ''
        for d in range(depth - 1, -2, -1):
            # get next result
            collection, subcollections, data_objects = next(colls)

            # check collection name
            if d >= 0:
                coll_name = 'sub' + str(d)
                self.assertEqual(collection.name, coll_name)
            else:
                # root collection
                self.assertEqual(collection.name, self.test_coll.name)

            # check subcollection name
            if d < depth - 1:
                self.assertEqual(sub_coll_name, subcollections[0].name)
            else:
                # last coll has no subcolls
                self.assertListEqual(subcollections, [])

            # check data object names
            for data_object in data_objects:
                self.assertIn(data_object.name, filenames)

            # iterate
            sub_coll_name = coll_name

        # that should be it
        with self.assertRaises(StopIteration):
            next(colls)

    def test_collection_metadata(self):
        self.assertIsInstance(self.test_coll.metadata, iRODSMetaCollection)


    def test_register_collection(self):
        tmp_dir = helpers.irods_shared_tmp_dir()
        loc_server = self.sess.host in ('localhost', socket.gethostname())
        if not(tmp_dir) and not(loc_server):
            self.skipTest('Requires access to server-side file(s)')

        # test vars
        file_count = 10
        dir_name = 'register_test_dir'
        dir_path = os.path.join((tmp_dir or '/tmp'), dir_name)
        coll_path = '{}/{}'.format(self.test_coll.path, dir_name)

        # make test dir
        helpers.make_flat_test_dir(dir_path, file_count)

        # register test dir
        self.sess.collections.register(dir_path, coll_path)

        # confirm collection presence
        coll = self.sess.collections.get(coll_path)

        # confirm object count in collection
        query = self.sess.query().count(DataObject.id).filter(Collection.name == coll_path)
        obj_count = next(query.get_results())[DataObject.id]
        self.assertEqual(file_count, int(obj_count))

        # remove coll but leave directory on disk
        coll.unregister()

        # delete test dir
        shutil.rmtree(dir_path)


    def test_register_collection_with_checksums(self):
        tmp_dir = helpers.irods_shared_tmp_dir()
        loc_server = self.sess.host in ('localhost', socket.gethostname())
        if not(tmp_dir) and not(loc_server):
            self.skipTest('Requires access to server-side file(s)')

        # test vars
        file_count = 10
        dir_name = 'register_test_dir_with_chksums'
        dir_path = os.path.join((tmp_dir or '/tmp'), dir_name)
        coll_path = '{}/{}'.format(self.test_coll.path, dir_name)

        # make test dir
        helpers.make_flat_test_dir(dir_path, file_count)

        # register test dir
        options = {kw.VERIFY_CHKSUM_KW: ''}
        self.sess.collections.register(dir_path, coll_path, **options)

        # confirm collection presence
        coll = self.sess.collections.get(coll_path)

        # confirm object count in collection
        query = self.sess.query().count(DataObject.id).filter(Collection.name == coll_path)
        obj_count = next(query.get_results())[DataObject.id]
        self.assertEqual(file_count, int(obj_count))

        # confirm object checksums
        objs = next(coll.walk())[2]
        for obj in objs:
            # don't use obj.path (aka logical path)
            phys_path = obj.replicas[0].path
            digest = helpers.compute_sha256_digest(phys_path)
            self.assertEqual(obj.checksum, "sha2:{}".format(digest))

        # remove coll but leave directory on disk
        coll.unregister()

        # delete test dir
        shutil.rmtree(dir_path)


if __name__ == "__main__":
    # let the tests find the parent irods lib
    sys.path.insert(0, os.path.abspath('../..'))
    unittest.main()



#!/usr/bin/env python
from __future__ import absolute_import
import os
import sys
import unittest

# this does not get called when imported from  runner.py
if __name__ == '__main__':
    sys.path.insert(0, os.path.abspath('../..'))

from irods.message import ET

# from base64 import b64encode, b64decode
from irods.message import (StartupPack, AuthResponse, IntegerIntegerMap,
                           IntegerStringMap, StringStringMap, GenQueryRequest,
                           GenQueryResponseColumn, GenQueryResponse)


class TestMessages(unittest.TestCase):

    def test_startup_pack(self):
        sup = StartupPack(('rods', 'tempZone'), ('rods', 'tempZone'))
        sup.irodsProt = 2
        sup.reconnFlag = 3
        sup.proxyUser = "rods"
        sup.proxyRcatZone = "tempZone"
        sup.clientUser = "rods"
        sup.clientRcatZone = "yoyoyo"
        sup.relVersion = "irods3.2"
        sup.apiVersion = "d"
        sup.option = "hellO"
        xml_str = sup.pack()
        expected = "<StartupPack_PI>\
<irodsProt>2</irodsProt>\
<reconnFlag>3</reconnFlag>\
<connectCnt>0</connectCnt>\
<proxyUser>rods</proxyUser>\
<proxyRcatZone>tempZone</proxyRcatZone>\
<clientUser>rods</clientUser>\
<clientRcatZone>yoyoyo</clientRcatZone>\
<relVersion>irods3.2</relVersion>\
<apiVersion>d</apiVersion>\
<option>hellO</option>\
</StartupPack_PI>"
        self.assertEqual(xml_str, expected)

        sup2 = StartupPack(('rods', 'tempZone'), ('rods', 'tempZone'))
        sup2.unpack(ET().fromstring(expected))
        self.assertEqual(sup2.irodsProt, 2)
        self.assertEqual(sup2.reconnFlag, 3)
        self.assertEqual(sup2.proxyUser, "rods")
        self.assertEqual(sup2.proxyRcatZone, "tempZone")
        self.assertEqual(sup2.clientUser, "rods")
        self.assertEqual(sup2.clientRcatZone, "yoyoyo")
        self.assertEqual(sup2.relVersion, "irods3.2")
        self.assertEqual(sup2.apiVersion, "d")
        self.assertEqual(sup2.option, "hellO")

    def test_auth_response(self):
        ar = AuthResponse()
        ar.response = b"hello"
        ar.username = "rods"
        expected = "<authResponseInp_PI>\
<response>aGVsbG8=</response>\
<username>rods</username>\
</authResponseInp_PI>"
        self.assertEqual(ar.pack(), expected)

        ar2 = AuthResponse()
        ar2.unpack(ET().fromstring(expected))
        self.assertEqual(ar2.response, b"hello")
        self.assertEqual(ar2.username, "rods")

    def test_inx_ival_pair(self):
        iip = IntegerIntegerMap()
        iip.iiLen = 2
        iip.inx = [4, 5]
        iip.ivalue = [1, 2]
        expected = "<InxIvalPair_PI>\
<iiLen>2</iiLen>\
<inx>4</inx>\
<inx>5</inx>\
<ivalue>1</ivalue>\
<ivalue>2</ivalue>\
</InxIvalPair_PI>"
        self.assertEqual(iip.pack(), expected)

        iip2 = IntegerIntegerMap()
        iip2.unpack(ET().fromstring(expected))
        self.assertEqual(iip2.iiLen, 2)
        self.assertEqual(iip2.inx, [4, 5])
        self.assertEqual(iip2.ivalue, [1, 2])

    def test_key_val_pair(self):
        kvp = StringStringMap()
        kvp.ssLen = 2
        kvp.keyWord = ["one", "two"]
        kvp.svalue = ["three", "four"]
        expected = "<KeyValPair_PI>\
<ssLen>2</ssLen>\
<keyWord>one</keyWord>\
<keyWord>two</keyWord>\
<svalue>three</svalue>\
<svalue>four</svalue>\
</KeyValPair_PI>"
        self.assertEqual(kvp.pack(), expected)

        kvp2 = StringStringMap()
        kvp2.unpack(ET().fromstring(expected))
        self.assertEqual(kvp2.ssLen, 2)
        self.assertEqual(kvp2.keyWord, ["one", "two"])
        self.assertEqual(kvp2.svalue, ["three", "four"])

    def test_gen_query_inp(self):
        gq = GenQueryRequest()
        gq.maxRows = 4
        gq.continueInx = 3
        gq.partialStartIndex = 2
        gq.options = 1

        kvp = StringStringMap()
        kvp.ssLen = 2
        kvp.keyWord = ['one', 'two']
        kvp.svalue = ['three', 'four']

        iip = IntegerIntegerMap()
        iip.iiLen = 2
        iip.inx = [4, 5]
        iip.ivalue = [1, 2]

        ivp = IntegerStringMap()
        ivp.isLen = 2
        ivp.inx = [1, 2]
        ivp.svalue = ['five', 'six']

        gq.KeyValPair_PI = kvp
        gq.InxIvalPair_PI = iip
        gq.InxValPair_PI = ivp

        expected = "<GenQueryInp_PI><maxRows>4</maxRows><continueInx>3</continueInx><partialStartIndex>2</partialStartIndex><options>1</options><KeyValPair_PI><ssLen>2</ssLen><keyWord>one</keyWord><keyWord>two</keyWord><svalue>three</svalue><svalue>four</svalue></KeyValPair_PI><InxIvalPair_PI><iiLen>2</iiLen><inx>4</inx><inx>5</inx><ivalue>1</ivalue><ivalue>2</ivalue></InxIvalPair_PI><InxValPair_PI><isLen>2</isLen><inx>1</inx><inx>2</inx><svalue>five</svalue><svalue>six</svalue></InxValPair_PI></GenQueryInp_PI>"
        self.assertEqual(gq.pack(), expected)

        gq2 = GenQueryRequest()
        gq2.unpack(ET().fromstring(expected))
        self.assertEqual(gq2.maxRows, 4)
        self.assertEqual(gq2.continueInx, 3)
        self.assertEqual(gq2.partialStartIndex, 2)
        self.assertEqual(gq2.options, 1)

        self.assertEqual(gq2.KeyValPair_PI.ssLen, 2)
        self.assertEqual(gq2.KeyValPair_PI.keyWord, ["one", "two"])
        self.assertEqual(gq2.KeyValPair_PI.svalue, ["three", "four"])

        self.assertEqual(gq2.InxIvalPair_PI.iiLen, 2)
        self.assertEqual(gq2.InxIvalPair_PI.inx, [4, 5])
        self.assertEqual(gq2.InxIvalPair_PI.ivalue, [1, 2])

        self.assertEqual(gq2.InxValPair_PI.isLen, 2)
        self.assertEqual(gq2.InxValPair_PI.inx, [1, 2])
        self.assertEqual(gq2.InxValPair_PI.svalue, ["five", "six"])

        self.assertEqual(gq2.pack(), expected)

    def test_sql_result(self):
        sr = GenQueryResponseColumn()
        sr.attriInx = 504
        sr.reslen = 64
        sr.value = ["one", "two"]

        expected = "<SqlResult_PI><attriInx>504</attriInx><reslen>64</reslen><value>one</value><value>two</value></SqlResult_PI>"
        self.assertEqual(sr.pack(), expected)

        sr2 = GenQueryResponseColumn()
        sr2.unpack(ET().fromstring(expected))
        self.assertEqual(sr2.attriInx, 504)
        self.assertEqual(sr2.reslen, 64)
        self.assertEqual(sr2.value, ["one", "two"])
        self.assertEqual(sr2.pack(), expected)

    def test_gen_query_out(self):
        gqo = GenQueryResponse()
        gqo.rowCnt = 2
        gqo.attriCnt = 2
        gqo.continueInx = 5
        gqo.totalRowCount = 7

        sr = GenQueryResponseColumn()
        sr.attriInx = 504
        sr.reslen = 64
        sr.value = ["one", "two"]
        gqo.SqlResult_PI = [sr, sr]

        expected = "<GenQueryOut_PI><rowCnt>2</rowCnt><attriCnt>2</attriCnt><continueInx>5</continueInx><totalRowCount>7</totalRowCount><SqlResult_PI><attriInx>504</attriInx><reslen>64</reslen><value>one</value><value>two</value></SqlResult_PI><SqlResult_PI><attriInx>504</attriInx><reslen>64</reslen><value>one</value><value>two</value></SqlResult_PI></GenQueryOut_PI>"
        self.assertEqual(gqo.pack(), expected)

        gqo2 = GenQueryResponse()
        gqo2.unpack(ET().fromstring(expected))

        self.assertEqual(gqo2.rowCnt, 2)
        self.assertEqual(gqo2.pack(), expected)

if __name__ == "__main__":
    unittest.main()



class iRODSAccount(object):

    def __init__(self, irods_host, irods_port, irods_user_name, irods_zone_name,
                 irods_authentication_scheme='native',
                 password=None, client_user=None,
                 server_dn=None, client_zone=None, **kwargs):

        self.authentication_scheme = irods_authentication_scheme.lower()
        self.host = irods_host
        self.port = int(irods_port)
        self.proxy_user = self.client_user = irods_user_name
        self.proxy_zone = self.client_zone = irods_zone_name
        self.server_dn = server_dn
        self.password = password

        for key, value in kwargs.items():
            try:
                if key.startswith('irods_'):
                    setattr(self, key[6:], value)
                else:
                    setattr(self, key, value)
            except TypeError:
                setattr(self, key, value)

        if client_user:
            self.client_user = client_user
            if client_zone:
                self.client_zone = client_zone



#! /usr/bin/env python
from __future__ import absolute_import
import os
import sys
import unittest

from irods.exception import OVERWRITE_WITHOUT_FORCE_FLAG
import irods.test.helpers as helpers

class TestForceCreate(unittest.TestCase):

    def setUp(self):
        self.sess = helpers.make_session()

    def tearDown(self):
        """Close connections."""
        self.sess.cleanup()

    # This test should pass whether or not federation is configured:
    def test_force_create(self):
        if self.sess.server_version > (4, 2, 8):
            self.skipTest('force flag unneeded for create in iRODS > 4.2.8')
        session = self.sess
        FILE = '/{session.zone}/home/{session.username}/a.txt'.format(**locals())
        try:
            session.data_objects.unlink(FILE)
        except:
            pass
        error = None
        try:
            session.data_objects.create(FILE)
            session.data_objects.create(FILE)
        except OVERWRITE_WITHOUT_FORCE_FLAG:
            error = "OVERWRITE_WITHOUT_FORCE_FLAG"
        self.assertEqual (error, "OVERWRITE_WITHOUT_FORCE_FLAG")
        error = None
        try:
            session.data_objects.create(FILE, force=True)
        except:
            error = "Error creating with force"
        self.assertEqual (error, None)
        try:
            session.data_objects.unlink(FILE)
        except:
            error = "Error cleaning up"
        self.assertEqual (error, None)


if __name__ == '__main__':
    # let the tests find the parent irods lib
    sys.path.insert(0, os.path.abspath('../..'))
    unittest.main()



from __future__ import absolute_import
import io
import sys
import logging
import six
import os
import ast

from irods.models import DataObject
from irods.meta import iRODSMetaCollection
import irods.keywords as kw
from irods.api_number import api_number
from irods.message import (JSON_Message, iRODSMessage)

logger = logging.getLogger(__name__)

IRODS_SERVER_WITH_CLOSE_REPLICA_API = (4,2,9)

def chunks(f, chunksize=io.DEFAULT_BUFFER_SIZE):
    return iter(lambda: f.read(chunksize), b'')

def irods_dirname(path):
    return path.rsplit('/', 1)[0]

def irods_basename(path):
    return path.rsplit('/', 1)[1]


class iRODSReplica(object):

    def __init__(self, number, status, resource_name, path, resc_hier, **kwargs):
        self.number = number
        self.status = status
        self.resource_name = resource_name
        self.path = path
        self.resc_hier = resc_hier
        for key, value in kwargs.items():
            setattr(self, key, value)


    def __repr__(self):
        return "<{}.{} {}>".format(
            self.__class__.__module__,
            self.__class__.__name__,
            self.resource_name
        )


class iRODSDataObject(object):

    def __init__(self, manager, parent=None, results=None):
        self.manager = manager
        if parent and results:
            self.collection = parent
            for attr, value in six.iteritems(DataObject.__dict__):
                if not attr.startswith('_'):
                    try:
                        setattr(self, attr, results[0][value])
                    except KeyError:
                        # backward compatibility with older schema versions
                        pass
            self.path = self.collection.path + '/' + self.name
            replicas = sorted(
                results, key=lambda r: r[DataObject.replica_number])
            self.replicas = [iRODSReplica(
                r[DataObject.replica_number],
                r[DataObject.replica_status],
                r[DataObject.resource_name],
                r[DataObject.path],
                r[DataObject.resc_hier],
                checksum=r[DataObject.checksum],
                size=r[DataObject.size]
            ) for r in replicas]
        self._meta = None




    def __repr__(self):
        return "<iRODSDataObject {id} {name}>".format(**vars(self))

    @property
    def metadata(self):
        if not self._meta:
            self._meta = iRODSMetaCollection(
                self.manager.sess.metadata, DataObject, self.path)
        return self._meta

    def open(self, mode='r', finalize_on_close = True, **options):
        return self.manager.open(self.path, mode, finalize_on_close = finalize_on_close, **options)

    def chksum(self, **options):
        """
        See: https://github.com/irods/irods/blob/4-2-stable/lib/api/include/dataObjChksum.h
        for a list of applicable irods.keywords options.

        NB options dict may also include a default-constructed RErrorStack object under the key r_error.
        If passed, this object can receive a list of warnings, one for each existing replica lacking a
        checksum.  (Relevant only in combination with VERIFY_CHKSUM_KW).
        """
        return self.manager.chksum(self.path, **options)

    def unlink(self, force=False, **options):
        self.manager.unlink(self.path, force, **options)

    def unregister(self, **options):
        self.manager.unregister(self.path, **options)

    def truncate(self, size):
        self.manager.truncate(self.path, size)

    def replicate(self, resource=None, **options):
        if resource:
            options[kw.DEST_RESC_NAME_KW] = resource
        self.manager.replicate(self.path, **options)


class iRODSDataObjectFileRaw(io.RawIOBase):

    """The raw object supporting file-like operations (read/write/seek) for the
       iRODSDataObject."""

    def __init__(self, conn, descriptor, finalize_on_close = True, **options):
        """
        Constructor needs a connection and an iRODS data object descriptor. If the
        finalize_on_close flag evaluates False, close() will invoke the REPLICA_CLOSE
        API instead of closing and finalizing the object (useful for parallel
        transfers using multiple threads).
        """
        super(iRODSDataObjectFileRaw,self).__init__()
        self.conn = conn
        self.desc = descriptor
        self.options = options
        self.finalize_on_close = finalize_on_close

    def replica_access_info(self):
        message_body = JSON_Message( {'fd': self.desc},
                                     server_version = self.conn.server_version )
        message = iRODSMessage('RODS_API_REQ', msg = message_body,
                               int_info=api_number['GET_FILE_DESCRIPTOR_INFO_APN'])
        self.conn.send(message)
        result = None
        try:
            result = self.conn.recv()
        except Exception as e:
            logger.warning('''Couldn't receive or process response to GET_FILE_DESCRIPTOR_INFO_APN -- '''
                           '''caught: %r''',e)
            raise
        dobj_info = result.get_json_encoded_struct()
        replica_token = dobj_info.get("replica_token","")
        resc_hier = ( dobj_info.get("data_object_info") or {} ).get("resource_hierarchy","")
        return (replica_token, resc_hier)

    def _close_replica(self):
        server_version = ast.literal_eval(os.environ.get('IRODS_VERSION_OVERRIDE', '()' ))
        if (server_version or self.conn.server_version) < IRODS_SERVER_WITH_CLOSE_REPLICA_API: return False
        message_body = JSON_Message( { "fd": self.desc,
                                       "send_notification": False,
                                       "update_size": False,
                                       "update_status": False,
                                       "compute_checksum": False },
                                     server_version = self.conn.server_version )
        self.conn.send( iRODSMessage('RODS_API_REQ', msg = message_body,
                                     int_info=api_number['REPLICA_CLOSE_APN']) )
        try:
            self.conn.recv().int_info
        except Exception:
            logger.warning ('** ERROR on closing replica **')
            raise
        return True

    def close(self):
        if self.finalize_on_close or not self._close_replica():
            self.conn.close_file(self.desc, **self.options)
        self.conn.release()
        super(iRODSDataObjectFileRaw, self).close()
        return None

    def seek(self, offset, whence=0):
        return self.conn.seek_file(self.desc, offset, whence)

    def readinto(self, b):
        contents = self.conn.read_file(self.desc, buffer=b)
        if contents is None:
            return 0

        return len(contents)

    def write(self, b):
        if isinstance(b, memoryview):
            return self.conn.write_file(self.desc, b.tobytes())

        return self.conn.write_file(self.desc, b)

    def readable(self):
        return True

    def writable(self):
        return True

    def seekable(self):
        return True



#!/usr/bin/env python

"""
NOTE: "If a test package name (directory with __init__.py) matches the pattern
       then the package will be checked for a load_tests function. If this
       exists then it will be called with loader, tests, pattern."
"""

from __future__ import absolute_import
import os
import sys
from unittest import TestLoader, TestSuite
import xmlrunner
import logging

logger = logging.getLogger()
logger.setLevel(logging.ERROR)
h = logging.StreamHandler()
f = logging.Formatter("%(asctime)s %(name)s-%(levelname)s [%(pathname)s %(lineno)d] %(message)s")
h.setFormatter(f)
logger.addHandler(h)


# Load all tests in the current directory and run them
if __name__ == "__main__":
    # must set the path for the imported tests
    sys.path.insert(0, os.path.abspath('../..'))

    loader = TestLoader()
    suite = TestSuite(loader.discover(start_dir='.', pattern='*_test.py',
                                      top_level_dir="."))

    result = xmlrunner.XMLTestRunner(
        verbosity=2, output='/tmp/python-irodsclient/test-reports').run(suite)
    if result.wasSuccessful():
        sys.exit(0)

    sys.exit(1)




# https://github.com/irods/irods/blob/4.2.1/lib/core/include/irods_client_server_negotiation.hpp
# https://github.com/irods/irods/blob/4.2.1/lib/core/src/irods_client_negotiation.cpp

# Token sent to the server to request negotiation
REQUEST_NEGOTIATION = "request_server_negotiation"

# Negotiation request values
REQUIRE_SSL = "CS_NEG_REQUIRE"
REQUIRE_TCP = "CS_NEG_REFUSE"

# Negotiation result (response) values
FAILURE = "CS_NEG_FAILURE"
USE_SSL = "CS_NEG_USE_SSL"
USE_TCP = "CS_NEG_USE_TCP"

# Keywords
CS_NEG_SID_KW = "cs_neg_sid_kw"
CS_NEG_RESULT_KW = "cs_neg_result_kw"


def perform_negotiation(client_policy, server_policy):
    if REQUIRE_SSL in (client_policy, server_policy):
        if REQUIRE_TCP in (client_policy, server_policy):
            return FAILURE, 0
        return USE_SSL, 1
    return USE_TCP, 1


def validate_policy(policy):
    if policy not in (REQUIRE_SSL, REQUIRE_TCP):
        raise ValueError('Invalid client-server negotiation policy: {}'.format(policy))



api_number = {
    # 500 - 599 - Internal File I/O API calls
    "FILE_CREATE_AN": 500,
    "FILE_OPEN_AN": 501,
    "FILE_WRITE_AN": 502,
    "FILE_CLOSE_AN": 503,
    "FILE_LSEEK_AN": 504,
    "FILE_READ_AN": 505,
    "FILE_UNLINK_AN": 506,
    "FILE_MKDIR_AN": 507,
    "FILE_CHMOD_AN": 508,
    "FILE_RMDIR_AN": 509,
    "FILE_STAT_AN": 510,
    "FILE_FSTAT_AN": 511,
    "FILE_FSYNC_AN": 512,
    "FILE_STAGE_AN": 513,
    "FILE_GET_FS_FREE_SPACE_AN": 514,
    "FILE_OPENDIR_AN": 515,
    "FILE_CLOSEDIR_AN": 516,
    "FILE_READDIR_AN": 517,
    "FILE_PUT_AN": 518,
    "FILE_GET_AN": 519,
    "FILE_CHKSUM_AN": 520,
    "CHK_N_V_PATH_PERM_AN": 521,
    "FILE_RENAME_AN": 522,
    "FILE_TRUNCATE_AN": 523,
    "FILE_STAGE_TO_CACHE_AN": 524,
    "FILE_SYNC_TO_ARCH_AN": 525,

    # 600 - 699 - Object File I/O API calls
    "DATA_OBJ_CREATE_AN": 601,
    "DATA_OBJ_OPEN_AN": 602,
    "DATA_OBJ_PUT_AN": 606,
    "DATA_PUT_AN": 607,
    "DATA_OBJ_GET_AN": 608,
    "DATA_GET_AN": 609,
    "DATA_OBJ_REPL250_AN": 610,
    "DATA_COPY_AN": 611,
    "DATA_OBJ_COPY250_AN": 613,
    "SIMPLE_QUERY_AN": 614,
    "DATA_OBJ_UNLINK_AN": 615,
    "REG_DATA_OBJ_AN": 619,
    "UNREG_DATA_OBJ_AN": 620,
    "REG_REPLICA_AN": 621,
    "MOD_DATA_OBJ_META_AN": 622,
    "RULE_EXEC_SUBMIT_AN": 623,
    "RULE_EXEC_DEL_AN": 624,
    "EXEC_MY_RULE_AN": 625,
    "OPR_COMPLETE_AN": 626,
    "DATA_OBJ_RENAME_AN": 627,
    "DATA_OBJ_RSYNC_AN": 628,
    "DATA_OBJ_CHKSUM_AN": 629,
    "PHY_PATH_REG_AN": 630,
    "DATA_OBJ_PHYMV250_AN": 631,
    "DATA_OBJ_TRIM_AN": 632,
    "OBJ_STAT_AN": 633,
    "SUB_STRUCT_FILE_CREATE_AN": 635,
    "SUB_STRUCT_FILE_OPEN_AN": 636,
    "SUB_STRUCT_FILE_READ_AN": 637,
    "SUB_STRUCT_FILE_WRITE_AN": 638,
    "SUB_STRUCT_FILE_CLOSE_AN": 639,
    "SUB_STRUCT_FILE_UNLINK_AN": 640,
    "SUB_STRUCT_FILE_STAT_AN": 641,
    "SUB_STRUCT_FILE_FSTAT_AN": 642,
    "SUB_STRUCT_FILE_LSEEK_AN": 643,
    "SUB_STRUCT_FILE_RENAME_AN": 644,
    "QUERY_SPEC_COLL_AN": 645,
    "SUB_STRUCT_FILE_MKDIR_AN": 647,
    "SUB_STRUCT_FILE_RMDIR_AN": 648,
    "SUB_STRUCT_FILE_OPENDIR_AN": 649,
    "SUB_STRUCT_FILE_READDIR_AN": 650,
    "SUB_STRUCT_FILE_CLOSEDIR_AN": 651,
    "DATA_OBJ_TRUNCATE_AN": 652,
    "SUB_STRUCT_FILE_TRUNCATE_AN": 653,
    "GET_XMSG_TICKET_AN": 654,
    "SEND_XMSG_AN": 655,
    "RCV_XMSG_AN": 656,
    "SUB_STRUCT_FILE_GET_AN": 657,
    "SUB_STRUCT_FILE_PUT_AN": 658,
    "SYNC_MOUNTED_COLL_AN": 659,
    "STRUCT_FILE_SYNC_AN": 660,
    "CLOSE_COLLECTION_AN": 661,
    "STRUCT_FILE_EXTRACT_AN": 664,
    "STRUCT_FILE_EXT_AND_REG_AN": 665,
    "STRUCT_FILE_BUNDLE_AN": 666,
    "CHK_OBJ_PERM_AND_STAT_AN": 667,
    "GET_REMOTE_ZONE_RESC_AN": 668,
    "DATA_OBJ_OPEN_AND_STAT_AN": 669,
    "L3_FILE_GET_SINGLE_BUF_AN": 670,
    "L3_FILE_PUT_SINGLE_BUF_AN": 671,
    "DATA_OBJ_CREATE_AND_STAT_AN": 672,
    "DATA_OBJ_CLOSE_AN": 673,
    "DATA_OBJ_LSEEK_AN": 674,
    "DATA_OBJ_READ_AN": 675,
    "DATA_OBJ_WRITE_AN": 676,
    "COLL_REPL_AN": 677,
    "OPEN_COLLECTION_AN": 678,
    "RM_COLL_AN": 679,
    "MOD_COLL_AN": 680,
    "COLL_CREATE_AN": 681,
    "RM_COLL_OLD_AN": 682,
    "REG_COLL_AN": 683,
    "PHY_BUNDLE_COLL_AN": 684,
    "UNBUN_AND_REG_PHY_BUNFILE_AN": 685,
    "GET_HOST_FOR_PUT_AN": 686,
    "GET_RESC_QUOTA_AN": 687,
    "BULK_DATA_OBJ_REG_AN": 688,
    "BULK_DATA_OBJ_PUT_AN": 689,
    "PROC_STAT_AN": 690,
    "STREAM_READ_AN": 691,
    "EXEC_CMD_AN": 692,
    "STREAM_CLOSE_AN": 693,
    "GET_HOST_FOR_GET_AN": 694,
    "DATA_OBJ_REPL_AN": 695,
    "DATA_OBJ_COPY_AN": 696,
    "DATA_OBJ_PHYMV_AN": 697,
    "DATA_OBJ_FSYNC_AN": 698,
    "DATA_OBJ_LOCK_AN": 699,

    # 700 - 799 - Metadata API calls
    "GET_MISC_SVR_INFO_AN": 700,
    "GENERAL_ADMIN_AN": 701,
    "GEN_QUERY_AN": 702,
    "AUTH_REQUEST_AN": 703,
    "AUTH_RESPONSE_AN": 704,
    "AUTH_CHECK_AN": 705,
    "MOD_AVU_METADATA_AN": 706,
    "MOD_ACCESS_CONTROL_AN": 707,
    "RULE_EXEC_MOD_AN": 708,
    "GET_TEMP_PASSWORD_AN": 709,
    "GENERAL_UPDATE_AN": 710,
    "GSI_AUTH_REQUEST_AN": 711,
    "READ_COLLECTION_AN": 713,
    "USER_ADMIN_AN": 714,
    "GENERAL_ROW_INSERT_AN": 715,
    "GENERAL_ROW_PURGE_AN": 716,
    "KRB_AUTH_REQUEST_AN": 717,
    "END_TRANSACTION_AN": 718,
    "DATABASE_RESC_OPEN_AN": 719,
    "DATABASE_OBJ_CONTROL_AN": 720,
    "DATABASE_RESC_CLOSE_AN": 721,
    "SPECIFIC_QUERY_AN": 722,
    "TICKET_ADMIN_AN": 723,
    "GET_TEMP_PASSWORD_FOR_OTHER_AN": 724,
    "PAM_AUTH_REQUEST_AN": 725,

    "EXEC_CMD241_AN": 634,

    "DATA_OBJ_READ201_AN": 603,
    "DATA_OBJ_WRITE201_AN": 604,
    "DATA_OBJ_CLOSE201_AN": 605,
    "DATA_OBJ_LSEEK201_AN": 612,
    "RM_COLL_OLD201_AN": 617,
    "REG_COLL201_AN": 618,
    "MOD_COLL201_AN": 646,
    "COLL_REPL201_AN": 662,
    "RM_COLL201_AN": 663,
    "OPEN_COLLECTION201_AN": 712,

    # 1000 - 1059 - NETCDF API calls
    "NC_OPEN_AN": 1000,
    "NC_CREATE_AN": 1001,
    "NC_CLOSE_AN": 1002,
    "NC_INQ_ID_AN": 1003,
    "NC_INQ_WITH_ID_AN": 1004,
    "NC_GET_VARS_BY_TYPE_AN": 1005,
    "NCCF_GET_VARA_AN": 1006,
    "NC_INQ_AN": 1007,
    "NC_OPEN_GROUP_AN": 1008,
    "NC_INQ_GRPS_AN": 1009,
    "NC_REG_GLOBAL_ATTR_AN": 1010,

    # 1060 - 1099 - OOI API calls
    "OOI_GEN_SERV_REQ_AN": 1060,

    # 1100 - 1200 - SSL API calls
    "SSL_START_AN": 1100,
    "SSL_END_AN": 1101,
    "ATOMIC_APPLY_METADATA_OPERATIONS_APN": 20002,
    "GET_FILE_DESCRIPTOR_INFO_APN": 20000,
    "REPLICA_CLOSE_APN": 20004
}



# Ordered property classes stolen from Kris Kowal of Ask a Wizard
# http://askawizard.blogspot.com/2008/10/ordered-properties-python-saga-part-5.html
from __future__ import absolute_import
from itertools import count
import six

try:
    next_counter = count().__next__
except AttributeError:
    next_counter = count().next


class OrderedProperty(object):

    def __init__(self, *args, **kws):
        self._creation_counter = next_counter()
        super(OrderedProperty, self).__init__(*args, **kws)


class OrderedMetaclass(type):

    def __init__(self, name, bases, attys):
        super(OrderedMetaclass, self).__init__(name, bases, attys)
        self._creation_counter = next_counter()
        self._ordered_properties = sorted(
            (
                (name, value)
                for base in reversed(self.__mro__)
                for name, value in base.__dict__.items()
                if isinstance(value, OrderedProperty)
                or isinstance(value, OrderedMetaclass)
            ),
            key=lambda property: property[1]._creation_counter,
        )


class OrderedClass(six.with_metaclass(OrderedMetaclass, object)):
    pass



#! /usr/bin/env python
from __future__ import absolute_import
import os
import sys
import unittest
from irods.exception import UserDoesNotExist
from irods.session import iRODSSession
import irods.test.helpers as helpers


class TestTempPassword(unittest.TestCase):
    """ Suite of tests for setting and getting temporary passwords as rodsadmin or rodsuser
    """
    admin = None
    new_user = 'bobby'
    password = 'foobar'

    @classmethod
    def setUpClass(cls):
        cls.admin = helpers.make_session()

    @classmethod
    def tearDownClass(cls):
        cls.admin.cleanup()

    def test_temp_password(self):
        # Make a new user
        self.admin.users.create(self.new_user, 'rodsuser')
        self.admin.users.modify(self.new_user, 'password', self.password)

        # Login as the new test user, to retrieve a temporary password
        with iRODSSession(host=self.admin.host,
                          port=self.admin.port,
                          user=self.new_user,
                          password=self.password,
                          zone=self.admin.zone) as session:
            # Obtain the temporary password
            conn = session.pool.get_connection()
            temp_password = conn.temp_password()

        # Open a new session with the temporary password
        with iRODSSession(host=self.admin.host,
                          port=self.admin.port,
                          user=self.new_user,
                          password=temp_password,
                          zone=self.admin.zone) as session:

            # do something that connects to the server
            session.users.get(self.admin.username)

        # delete new user
        self.admin.users.remove(self.new_user)

        # user should be gone
        with self.assertRaises(UserDoesNotExist):
            self.admin.users.get(self.new_user)

    def test_set_temp_password(self):
        # make a new user
        temp_user = self.admin.users.create(self.new_user, 'rodsuser')

        # obtain a temporary password as rodsadmin for another user
        temp_password = temp_user.temp_password()

        # open a session as the new user
        with iRODSSession(host=self.admin.host,
                          port=self.admin.port,
                          user=self.new_user,
                          password=temp_password,
                          zone=self.admin.zone) as session:

            # do something that connects to the server
            session.users.get(self.new_user)

        # delete new user
        self.admin.users.remove(self.new_user)

        # user should be gone
        with self.assertRaises(UserDoesNotExist):
            self.admin.users.get(self.new_user)


if __name__ == '__main__':
    # let the tests find the parent irods lib
    sys.path.insert(0, os.path.abspath('../..'))
    unittest.main()



from __future__ import absolute_import
from irods.models import Zone


class iRODSZone(object):

    def __init__(self, manager, result=None):
        """Construct an iRODSZone object."""
        self.manager = manager
        if result:
            self.id = result[Zone.id]
            self.name = result[Zone.name]
            self.type = result[Zone.type]

    def remove(self):
        self.manager.remove(self.name)

    def __repr__(self):
        """Render a user-friendly string representation for the iRODSZone object."""
        return "<iRODSZone {id} {name} {type}>".format(**vars(self))




from __future__ import absolute_import
import itertools
import operator

from irods.models import Collection, DataObject
from irods.data_object import iRODSDataObject, irods_basename
from irods.meta import iRODSMetaCollection


class iRODSCollection(object):

    def __init__(self, manager, result=None):
        self.manager = manager
        if result:
            self.id = result[Collection.id]
            self.path = result[Collection.name]
            self.name = irods_basename(result[Collection.name])
            self.create_time = result[Collection.create_time]
            self.modify_time = result[Collection.modify_time]
            self._inheritance = result[Collection.inheritance]
            self.owner_name = result[Collection.owner_name]
            self.owner_zone = result[Collection.owner_zone]
        self._meta = None

    @property
    def inheritance(self):
        return bool(self._inheritance) and self._inheritance != "0"

    @property
    def metadata(self):
        if not self._meta:
            self._meta = iRODSMetaCollection(self.manager.sess.metadata,
                                             Collection, self.path)
        return self._meta

    @property
    def subcollections(self):
        query = self.manager.sess.query(Collection)\
            .filter(Collection.parent_name == self.path)
        return [iRODSCollection(self.manager, row) for row in query]

    @property
    def data_objects(self):
        query = self.manager.sess.query(DataObject)\
            .filter(Collection.name == self.path)
        results = query.get_results()
        grouped = itertools.groupby(
            results, operator.itemgetter(DataObject.id))
        return [
            iRODSDataObject(
                self.manager.sess.data_objects, self, list(replicas))
            for _, replicas in grouped
        ]

    def remove(self, recurse=True, force=False, **options):
        self.manager.remove(self.path, recurse, force, **options)

    def unregister(self, **options):
        self.manager.unregister(self.path, **options)

    def move(self, path):
        self.manager.move(self.path, path)

    def walk(self, topdown=True):
        """
        Collection tree generator.

        For each subcollection in the directory tree, starting at the
        collection, yield a 3-tuple
        """

        if topdown:
            yield (self, self.subcollections, self.data_objects)
        for subcollection in self.subcollections:
            new_root = subcollection
            for x in new_root.walk(topdown):
                yield x
        if not topdown:
            yield (self, self.subcollections, self.data_objects)

    def __repr__(self):
        return "<iRODSCollection {id} {name}>".format(id=self.id, name=self.name.encode('utf-8'))



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
import os
import sys
import unittest
from irods.meta import (iRODSMeta, AVUOperation, BadAVUOperationValue, BadAVUOperationKeyword)
from irods.manager.metadata_manager import InvalidAtomicAVURequest
from irods.models import (DataObject, Collection, Resource)
import irods.test.helpers as helpers
from six.moves import range
from six import PY3


class TestMeta(unittest.TestCase):
    '''Suite of tests on metadata operations
    '''
    # test metadata
    attr0, value0, unit0 = 'attr0', 'value0', 'unit0'
    attr1, value1, unit1 = 'attr1', 'value1', 'unit1'

    def setUp(self):
        self.sess = helpers.make_session()
        # test data
        self.coll_path = '/{}/home/{}/test_dir'.format(self.sess.zone, self.sess.username)
        self.obj_name = 'test1'
        self.obj_path = '{coll_path}/{obj_name}'.format(**vars(self))

        # Create test collection and (empty) test object
        self.coll = self.sess.collections.create(self.coll_path)
        self.obj = self.sess.data_objects.create(self.obj_path)

    def tearDown(self):
        '''Remove test data and close connections
        '''
        self.coll.remove(recurse=True, force=True)
        helpers.remove_unused_metadata(self.sess)
        self.sess.cleanup()

    from irods.test.helpers import create_simple_resc_hierarchy

    def test_atomic_metadata_operations_244(self):
        user = self.sess.users.get("rods")
        group = self.sess.user_groups.get("public")
        m = ( "attr_244","value","units")

        with self.assertRaises(BadAVUOperationValue):
            AVUOperation(operation="add", avu=m)

        with self.assertRaises(BadAVUOperationValue):
            AVUOperation(operation="not_add_or_remove", avu=iRODSMeta(*m))

        with self.assertRaises(BadAVUOperationKeyword):
            AVUOperation(operation="add", avu=iRODSMeta(*m), extra_keyword=None)


        with self.assertRaises(InvalidAtomicAVURequest):
            user.metadata.apply_atomic_operations( tuple() )

        user.metadata.apply_atomic_operations()   # no AVUs applied - no-op without error

        for n,obj in enumerate((group, user, self.coll, self.obj)):
            avus = [ iRODSMeta('some_attribute',str(i),'some_units') for i in range(n*100,(n+1)*100) ]
            obj.metadata.apply_atomic_operations(*[AVUOperation(operation="add", avu=avu_) for avu_ in avus])
            obj.metadata.apply_atomic_operations(*[AVUOperation(operation="remove", avu=avu_) for avu_ in avus])


    def test_atomic_metadata_operation_for_resource_244(self):
        (root,leaf)=('ptX','rescX')
        with self.create_simple_resc_hierarchy(root,leaf):
            root_resc = self.sess.resources.get(root)   # resource objects
            leaf_resc = self.sess.resources.get(leaf)
            root_tuple = ('role','root','new units #1')    # AVU tuples to apply
            leaf_tuple = ('role','leaf','new units #2')
            root_resc.metadata.add( *root_tuple[:2] ) # first apply without units ...
            leaf_resc.metadata.add( *leaf_tuple[:2] )
            for resc,resc_tuple in ((root_resc, root_tuple), (leaf_resc, leaf_tuple)):
                resc.metadata.apply_atomic_operations(  # metadata set operation (remove + add) to add units
                    AVUOperation(operation="remove", avu=iRODSMeta(*resc_tuple[:2])),
                    AVUOperation(operation="add", avu=iRODSMeta(*resc_tuple[:3]))
                )
                resc_meta = self.sess.metadata.get(Resource, resc.name)
                avus_to_tuples = lambda avu_list: sorted([(i.name,i.value,i.units) for i in avu_list])
                self.assertEqual(avus_to_tuples(resc_meta), avus_to_tuples([iRODSMeta(*resc_tuple)]))


    def test_atomic_metadata_operation_for_data_object_244(self):
        AVUs_Equal = lambda avu1,avu2,fn=(lambda x:x): fn(avu1)==fn(avu2)
        AVU_As_Tuple = lambda avu,length=3:(avu.name,avu.value,avu.units)[:length]
        AVU_Units_String = lambda avu:"" if not avu.units else avu.units
        m = iRODSMeta( "attr_244","value","units")
        self.obj.metadata.add(m)
        meta = self.sess.metadata.get(DataObject, self.obj_path)
        self.assertEqual(len(meta), 1)
        self.assertTrue(AVUs_Equal(m,meta[0],AVU_As_Tuple))
        self.obj.metadata.apply_atomic_operations(                                  # remove original AVU and replace
           AVUOperation(operation="remove",avu=m),                                  #   with two altered versions
           AVUOperation(operation="add",avu=iRODSMeta(m.name,m.value,"units_244")), # (one of them without units) ...
           AVUOperation(operation="add",avu=iRODSMeta(m.name,m.value))
        )
        meta = self.sess.metadata.get(DataObject, self.obj_path)   # ... check integrity of change
        self.assertEqual(sorted([AVU_Units_String(i) for i in meta]), ["","units_244"])

    def test_atomic_metadata_operations_255(self):
        my_resc = self.sess.resources.create('dummyResc','passthru')
        avus = [iRODSMeta('a','b','c'), iRODSMeta('d','e','f')]
        objects = [ self.sess.users.get("rods"), self.sess.user_groups.get("public"), my_resc,
                    self.sess.collections.get(self.coll_path), self.sess.data_objects.get(self.obj_path)  ]
        try:
            for obj in objects:
                self.assertEqual(len(obj.metadata.items()), 0)
                for n,item in enumerate(avus):
                    obj.metadata.apply_atomic_operations(AVUOperation(operation='add',avu=item))
                    self.assertEqual(len(obj.metadata.items()), n+1)
        finally:
            for obj in objects: obj.metadata.remove_all()
            my_resc.remove()

    def test_get_obj_meta(self):
        # get object metadata
        meta = self.sess.metadata.get(DataObject, self.obj_path)

        # there should be no metadata at this point
        assert len(meta) == 0

    def test_resc_meta(self):
        rescname = 'demoResc'
        self.sess.resources.get(rescname).metadata.remove_all()
        self.sess.metadata.set(Resource, rescname, iRODSMeta('zero','marginal','cost'))
        self.sess.metadata.add(Resource, rescname, iRODSMeta('zero','marginal'))
        self.sess.metadata.set(Resource, rescname, iRODSMeta('for','ever','after'))
        meta = self.sess.resources.get(rescname).metadata
        self.assertTrue( len(meta) == 3 )
        resource = self.sess.resources.get(rescname)
        all_AVUs= resource.metadata.items()
        for avu in all_AVUs:
            resource.metadata.remove(avu)
        self.assertTrue(0 == len(self.sess.resources.get(rescname).metadata))

    def test_add_obj_meta(self):
        # add metadata to test object
        self.sess.metadata.add(DataObject, self.obj_path,
                               iRODSMeta(self.attr0, self.value0))
        self.sess.metadata.add(DataObject, self.obj_path,
                               iRODSMeta(self.attr1, self.value1, self.unit1))

        # Throw in some unicode for good measure
        attribute, value = 'attr2', u'☭⛷★⚽'
        self.sess.metadata.add(DataObject, self.obj_path,
                               iRODSMeta(attribute, value))

        # get object metadata
        meta = self.sess.metadata.get(DataObject, self.obj_path)

        # sort results by metadata id
        def getKey(AVU):
            return AVU.avu_id
        meta = sorted(meta, key=getKey)

        # assertions
        assert meta[0].name == self.attr0
        assert meta[0].value == self.value0

        assert meta[1].name == self.attr1
        assert meta[1].value == self.value1
        assert meta[1].units == self.unit1

        assert meta[2].name == attribute
        testValue = (value if PY3 else value.encode('utf8'))
        assert meta[2].value == testValue


    def test_add_obj_meta_empty(self):
        '''Should raise exception
        '''

        # try to add metadata with empty value
        with self.assertRaises(ValueError):
            self.sess.metadata.add(DataObject, self.obj_path,
                                   iRODSMeta('attr_with_empty_value', ''))


    def test_copy_obj_meta(self):
        # test destination object for copy
        dest_obj_path = self.coll_path + '/test2'
        self.sess.data_objects.create(dest_obj_path)

        # add metadata to test object
        self.sess.metadata.add(DataObject, self.obj_path,
                               iRODSMeta(self.attr0, self.value0))

        # copy metadata
        self.sess.metadata.copy(
            DataObject, DataObject, self.obj_path, dest_obj_path)

        # get destination object metadata
        dest_meta = self.sess.metadata.get(DataObject, dest_obj_path)

        # check metadata
        assert dest_meta[0].name == self.attr0


    def test_remove_obj_meta(self):
        # add metadata to test object
        self.sess.metadata.add(DataObject, self.obj_path,
                               iRODSMeta(self.attr0, self.value0))

        # check that metadata is there
        meta = self.sess.metadata.get(DataObject, self.obj_path)
        assert meta[0].name == self.attr0

        # remove metadata from object
        self.sess.metadata.remove(DataObject, self.obj_path,
                                  iRODSMeta(self.attr0, self.value0))

        # check that metadata is gone
        meta = self.sess.metadata.get(DataObject, self.obj_path)
        assert len(meta) == 0


    def test_add_coll_meta(self):
        # add metadata to test collection
        self.sess.metadata.add(Collection, self.coll_path,
                               iRODSMeta(self.attr0, self.value0))

        # get collection metadata
        meta = self.sess.metadata.get(Collection, self.coll_path)

        # assertions
        assert meta[0].name == self.attr0
        assert meta[0].value == self.value0

        # remove collection metadata
        self.sess.metadata.remove(Collection, self.coll_path,
                                  iRODSMeta(self.attr0, self.value0))

        # check that metadata is gone
        meta = self.sess.metadata.get(Collection, self.coll_path)
        assert len(meta) == 0


    def test_meta_repr(self):
        # test obj
        collection = self.coll_path
        filename = 'test_meta_repr.txt'
        test_obj_path = '{collection}/{filename}'.format(**locals())

        # make object
        obj = helpers.make_object(self.sess, test_obj_path)

        # test AVU
        attribute, value, units = ('test_attr', 'test_value', 'test_units')

        # add metadata to test object
        meta = self.sess.metadata.add(DataObject, test_obj_path,
                                      iRODSMeta(attribute, value, units))

        # get metadata
        meta = self.sess.metadata.get(DataObject, test_obj_path)
        avu_id = meta[0].avu_id

        # assert
        self.assertEqual(
            repr(meta[0]), "<iRODSMeta {avu_id} {attribute} {value} {units}>".format(**locals()))

        # remove test object
        obj.unlink(force=True)


    def test_irodsmetacollection_data_obj(self):
        '''
        Tested as data_object metadata
        '''
        # test settings
        avu_count = 5

        # make test object
        test_obj_path = self.coll_path + '/test_irodsmetacollection'
        test_obj = helpers.make_object(self.sess, test_obj_path)

        # test AVUs
        triplets = [('test_attr' + str(i), 'test_value', 'test_units')
                    for i in range(avu_count)]

        # get coll meta
        imc = test_obj.metadata

        # try invalid key
        with self.assertRaises(KeyError):
            imc.get_one('bad_key')

        # invalid key type
        with self.assertRaises(TypeError):
            imc.get_one(list())

        # try empty update values
        with self.assertRaises(ValueError):
            imc.add()

        # add AVUs
        for triplet in triplets:
            imc.add(*triplet)

        # add another AVU with existing attribute name
        attr_name = triplets[0][0]
        duplicate_triplet = (attr_name, 'other_value', 'test_units')
        imc.add(*duplicate_triplet)

        # get_one should fail
        with self.assertRaises(KeyError):
            imc.get_one(attr_name)

        # remove triplet
        imc.remove(*duplicate_triplet)
        imc.get_one(attr_name)

        # get keys
        for key in imc.keys():
            self.assertIn(key, [triplet[0] for triplet in triplets])

        # get items
        for avu in imc.items():
            self.assertIsInstance(avu, iRODSMeta)
            self.assertIn(avu.name, [triplet[0] for triplet in triplets])
            self.assertIn(avu.value, [triplet[1] for triplet in triplets])
            self.assertIn(avu.units, [triplet[2] for triplet in triplets])

        # try contains
        self.assertIn(triplets[0][0], imc)

        # try contains with bad key type
        with self.assertRaises(TypeError):
            _ = (int() in imc)

        # set item
        imc[attr_name] = iRODSMeta(attr_name, 'boo')

        # get item
        _ = imc[attr_name]

        # del item with bad key type
        with self.assertRaises(TypeError):
            del imc[int()]

        # del item
        del imc[attr_name]

        with self.assertRaises(KeyError):
            _ = imc[attr_name]

        # remove all metadta
        imc.remove_all()
        self.assertEqual(len(imc), 0)

        # remove test collection
        test_obj.unlink(force=True)


if __name__ == '__main__':
    # let the tests find the parent irods lib
    sys.path.insert(0, os.path.abspath('../..'))
    unittest.main()



'''From rodsKeyWdDef.hpp
'''

ALL_KW = "all"        # operation done on all replicas #
COPIES_KW = "copies"    # the number of copies #
EXEC_LOCALLY_KW = "execLocally"    # execute locally #
FORCE_FLAG_KW = "forceFlag"    # force update #
CLI_IN_SVR_FIREWALL_KW = "cliInSvrFirewall"  # cli behind same firewall #
REG_CHKSUM_KW = "regChksum"    # register checksum #
VERIFY_CHKSUM_KW = "verifyChksum"   # verify checksum #
NO_COMPUTE_KW = "no_compute"
VERIFY_BY_SIZE_KW = "verifyBySize"  # verify by size - used by irsync #
OBJ_PATH_KW = "objPath"   # logical path of the object #
RESC_NAME_KW = "rescName"   # resource name #
DEST_RESC_NAME_KW = "destRescName"   # destination resource name #
DEF_RESC_NAME_KW = "defRescName"   # default resource name #
BACKUP_RESC_NAME_KW = "backupRescName"  # backup resource name #
DATA_TYPE_KW = "dataType"   # data type #
DATA_SIZE_KW = "dataSize"
CHKSUM_KW = "chksum"
ORIG_CHKSUM_KW = "orig_chksum"
VERSION_KW = "version"
FILE_PATH_KW = "filePath"   # the physical file path #
BUN_FILE_PATH_KW = "bunFilePath"  # the physical bun file path # # JMC - backport 4768
REPL_NUM_KW = "replNum"   # replica number #
WRITE_FLAG_KW = "writeFlag"   # whether it is opened for write #
REPL_STATUS_KW = "replStatus"   # status of the replica #
ALL_REPL_STATUS_KW = "allReplStatus"   # update all replStatus #
METADATA_INCLUDED_KW = "metadataIncluded" # for atomic puts of data / metadata
ACL_INCLUDED_KW = "aclIncluded" # for atomic puts of data / access controls
DATA_INCLUDED_KW = "dataIncluded"   # data included in the input buffer #
DATA_OWNER_KW = "dataOwner"
DATA_OWNER_ZONE_KW = "dataOwnerZone"
DATA_EXPIRY_KW = "dataExpiry"
DATA_COMMENTS_KW = "dataComments"
DATA_CREATE_KW = "dataCreate"
DATA_MODIFY_KW = "dataModify"
DATA_ACCESS_KW = "dataAccess"
DATA_ACCESS_INX_KW = "dataAccessInx"
NO_OPEN_FLAG_KW = "noOpenFlag"
PHYOPEN_BY_SIZE_KW = "phyOpenBySize"
STREAMING_KW = "streaming"
DATA_ID_KW = "dataId"
COLL_ID_KW = "collId"
DATA_MODE_KW = "dataMode"
STATUS_STRING_KW = "statusString"
DATA_MAP_ID_KW = "dataMapId"
NO_PARA_OP_KW = "noParaOpr"
LOCAL_PATH_KW = "localPath"
RSYNC_MODE_KW = "rsyncMode"
RSYNC_DEST_PATH_KW = "rsyncDestPath"
RSYNC_CHKSUM_KW = "rsyncChksum"
CHKSUM_ALL_KW = "ChksumAll"
FORCE_CHKSUM_KW = "forceChksum"
COLLECTION_KW = "collection"
ADMIN_KW = "irodsAdmin"
ADMIN_RMTRASH_KW = "irodsAdminRmTrash"
UNREG_KW = "unreg"
RMTRASH_KW = "irodsRmTrash"
RECURSIVE_OPR__KW = "recursiveOpr"
COLLECTION_TYPE_KW = "collectionType"
COLLECTION_INFO1_KW = "collectionInfo1"
COLLECTION_INFO2_KW = "collectionInfo2"
SEL_OBJ_TYPE_KW = "selObjType"
STRUCT_FILE_OPR_KW = "structFileOpr"
ALL_MS_PARAM_KW = "allMsParam"
UNREG_COLL_KW = "unregColl"
UPDATE_REPL_KW = "updateRepl"
RBUDP_TRANSFER_KW = "rbudpTransfer"
VERY_VERBOSE_KW = "veryVerbose"
RBUDP_SEND_RATE_KW = "rbudpSendRate"
RBUDP_PACK_SIZE_KW = "rbudpPackSize"
ZONE_KW = "zone"
REMOTE_ZONE_OPR_KW = "remoteZoneOpr"
REPL_DATA_OBJ_INP_KW = "replDataObjInp"
CROSS_ZONE_CREATE_KW = "replDataObjInp"  # use the same for backward compatibility #
QUERY_BY_DATA_ID_KW = "queryByDataID"
SU_CLIENT_USER_KW = "suClientUser"
RM_BUN_COPY_KW = "rmBunCopy"
KEY_WORD_KW = "keyWord"  # the msKeyValStr is a keyword #
CREATE_MODE_KW = "createMode"  # a msKeyValStr keyword #
OPEN_FLAGS_KW = "openFlags"  # a msKeyValStr keyword #
OFFSET_KW = "offset"  # a msKeyValStr keyword #
# DATA_SIZE_KW already defined #
NUM_THREADS_KW = "numThreads"  # a msKeyValStr keyword #
OPR_TYPE_KW = "oprType"  # a msKeyValStr keyword #
COLL_FLAGS_KW = "collFlags"  # a msKeyValStr keyword #
TRANSLATED_PATH_KW = "translatedPath"  # the path translated #
NO_TRANSLATE_LINKPT_KW = "noTranslateMntpt"  # don't translate mntpt #
BULK_OPR_KW = "bulkOpr"  # the bulk operation #
NON_BULK_OPR_KW = "nonBulkOpr"  # non bulk operation #
EXEC_CMD_RULE_KW = "execCmdRule"  # the rule that invoke execCmd #
EXEC_MY_RULE_KW = "execMyRule"  # the rule is invoked by rsExecMyRule #
STREAM_STDOUT_KW = "streamStdout"  # the stream stdout for execCmd #
REG_REPL_KW = "regRepl"  # register replica #
AGE_KW = "age"  # age of the file for itrim #
DRYRUN_KW = "dryrun"  # do a dry run #
ACL_COLLECTION_KW = "aclCollection"  # the collection from which the ACL should be used #
NO_CHK_COPY_LEN_KW = "noChkCopyLen"  # Don't check the len when transfering  #
TICKET_KW = "ticket"       # for ticket-based-access #
PURGE_CACHE_KW = "purgeCache"   # purge the cache copy right after the operation JMC - backport 4537
EMPTY_BUNDLE_ONLY_KW = "emptyBundleOnly"  # delete emptyBundleOnly # # JMC - backport 4552

# =-=-=-=-=-=-=-
# JMC - backport 4599
LOCK_TYPE_KW = "lockType"     # valid values are READ_LOCK_TYPE, WRITE_LOCK_TYPE and UNLOCK_TYPE #
LOCK_CMD_KW = "lockCmd"      # valid values are SET_LOCK_WAIT_CMD, SET_LOCK_CMD and GET_LOCK_CMD #
LOCK_FD_KW = "lockFd"       # Lock file desc for unlock #
MAX_SUB_FILE_KW = "maxSubFile"  # max number of files for tar file bundles #
MAX_BUNDLE_SIZE_KW = "maxBunSize"  # max size of a tar bundle in Gbs #
NO_STAGING_KW = "noStaging"

# =-=-=-=-=-=-=-
MAX_SUB_FILE_KW = "maxSubFile"  # max number of files for tar file bundles # # JMC - backport 4771

# OBJ_PATH_KW already defined #

# OBJ_PATH_KW already defined #
# COLL_NAME_KW already defined #
FILE_UID_KW = "fileUid"
FILE_OWNER_KW = "fileOwner"
FILE_GID_KW = "fileGid"
FILE_GROUP_KW = "fileGroup"
FILE_MODE_KW = "fileMode"
FILE_CTIME_KW = "fileCtime"
FILE_MTIME_KW = "fileMtime"
FILE_SOURCE_PATH_KW = "fileSourcePath"
EXCLUDE_FILE_KW = "excludeFile"

# The following are the keyWord definition for the rescCond key/value pair #
# RESC_NAME_KW is defined above #

RESC_ZONE_KW = "zoneName"
RESC_LOC_KW = "rescLoc"  # resc_net in DB #
RESC_TYPE_KW = "rescType"
RESC_CLASS_KW = "rescClass"
RESC_VAULT_PATH_KW = "rescVaultPath"  # resc_def_path in DB #
RESC_STATUS_KW = "rescStatus"
GATEWAY_ADDR_KW = "gateWayAddr"
RESC_MAX_OBJ_SIZE_KW = "rescMaxObjSize"
FREE_SPACE_KW = "freeSpace"
FREE_SPACE_TIME_KW = "freeSpaceTime"
FREE_SPACE_TIMESTAMP_KW = "freeSpaceTimeStamp"
RESC_TYPE_INX_KW = "rescTypeInx"
RESC_CLASS_INX_KW = "rescClassInx"
RESC_ID_KW = "rescId"
RESC_COMMENTS_KW = "rescComments"
RESC_CREATE_KW = "rescCreate"
RESC_MODIFY_KW = "rescModify"

# The following are the keyWord definition for the userCond key/value pair #

USER_NAME_CLIENT_KW = "userNameClient"
RODS_ZONE_CLIENT_KW = "rodsZoneClient"
HOST_CLIENT_KW = "hostClient"
CLIENT_ADDR_KW = "clientAddr"
USER_TYPE_CLIENT_KW = "userTypeClient"
AUTH_STR_CLIENT_KW = "authStrClient"  # user distin name #
USER_AUTH_SCHEME_CLIENT_KW = "userAuthSchemeClient"
USER_INFO_CLIENT_KW = "userInfoClient"
USER_COMMENT_CLIENT_KW = "userCommentClient"
USER_CREATE_CLIENT_KW = "userCreateClient"
USER_MODIFY_CLIENT_KW = "userModifyClient"
USER_NAME_PROXY_KW = "userNameProxy"
RODS_ZONE_PROXY_KW = "rodsZoneProxy"
HOST_PROXY_KW = "hostProxy"
USER_TYPE_PROXY_KW = "userTypeProxy"
AUTH_STR_PROXY_KW = "authStrProxy"  # dn #
USER_AUTH_SCHEME_PROXY_KW = "userAuthSchemeProxy"
USER_INFO_PROXY_KW = "userInfoProxy"
USER_COMMENT_PROXY_KW = "userCommentProxy"
USER_CREATE_PROXY_KW = "userCreateProxy"
USER_MODIFY_PROXY_KW = "userModifyProxy"
ACCESS_PERMISSION_KW = "accessPermission"
NO_CHK_FILE_PERM_KW = "noChkFilePerm"

# The following are the keyWord definition for the collCond key/value pair #

COLL_NAME_KW = "collName"
COLL_PARENT_NAME_KW = "collParentName"  # parent_coll_name in DB  #
COLL_OWNER_NAME_KW = "collOwnername"
COLL_OWNER_ZONE_KW = "collOwnerZone"
COLL_MAP_ID_KW = "collMapId"
COLL_INHERITANCE_KW = "collInheritance"
COLL_COMMENTS_KW = "collComments"
COLL_EXPIRY_KW = "collExpiry"
COLL_CREATE_KW = "collCreate"
COLL_MODIFY_KW = "collModify"
COLL_ACCESS_KW = "collAccess"
COLL_ACCESS_INX_KW = "collAccessInx"
COLL_ID_KW = "collId"

# The following are the keyWord definitions for the keyValPair_t input to chlModRuleExec.
#
RULE_NAME_KW = "ruleName"
RULE_REI_FILE_PATH_KW = "reiFilePath"
RULE_USER_NAME_KW = "userName"
RULE_EXE_ADDRESS_KW = "exeAddress"
RULE_EXE_TIME_KW = "exeTime"
RULE_EXE_FREQUENCY_KW = "exeFrequency"
RULE_PRIORITY_KW = "priority"
RULE_ESTIMATE_EXE_TIME_KW = "estimateExeTime"
RULE_NOTIFICATION_ADDR_KW = "notificationAddr"
RULE_LAST_EXE_TIME_KW = "lastExeTime"
RULE_EXE_STATUS_KW = "exeStatus"


EXCLUDE_FILE_KW = "excludeFile"
AGE_KW = "age"  # age of the file for itrim #

# =-=-=-=-=-=-=-
# irods general keywords definitions
RESC_HIER_STR_KW = "resc_hier"
REPLICA_TOKEN_KW = "replicaToken"
DEST_RESC_HIER_STR_KW = "dest_resc_hier"
IN_PDMO_KW = "in_pdmo"
STAGE_OBJ_KW = "stage_object"
SYNC_OBJ_KW = "sync_object"
IN_REPL_KW = "in_repl"

# =-=-=-=-=-=-=-
# irods tcp keyword definitions
SOCKET_HANDLE_KW = "tcp_socket_handle"

# =-=-=-=-=-=-=-
# irods ssl keyword definitions
SSL_HOST_KW = "ssl_host"
SSL_SHARED_SECRET_KW = "ssl_shared_secret"
SSL_KEY_SIZE_KW = "ssl_key_size"
SSL_SALT_SIZE_KW = "ssl_salt_size"
SSL_NUM_HASH_ROUNDS_KW = "ssl_num_hash_rounds"
SSL_ALGORITHM_KW = "ssl_algorithm"

# =-=-=-=-=-=-=-
# irods data_object keyword definitions
PHYSICAL_PATH_KW = "physical_path"
MODE_KW = "mode_kw"
FLAGS_KW = "flags_kw"
# borrowed RESC_HIER_STR_KW

# =-=-=-=-=-=-=-
# irods file_object keyword definitions
LOGICAL_PATH_KW = "logical_path"
FILE_DESCRIPTOR_KW = "file_descriptor"
L1_DESC_IDX_KW = "l1_desc_idx"
SIZE_KW = "file_size"
REPL_REQUESTED_KW = "repl_requested"
# borrowed IN_PDMO_KW

# =-=-=-=-=-=-=-
# irods structured_object keyword definitions
HOST_ADDR_KW = "host_addr"
ZONE_NAME_KW = "zone_name"
PORT_NUM_KW = "port_num"
SUB_FILE_PATH_KW = "sub_file_path"
# borrowed OFFSET_KW
# borrowed DATA_TYPE_KW
# borrowed OPR_TYPE_KW

# =-=-=-=-=-=-=-
# irods spec coll keyword definitions
SPEC_COLL_CLASS_KW = "spec_coll_class"
SPEC_COLL_TYPE_KW = "spec_coll_type"
SPEC_COLL_OBJ_PATH_KW = "spec_coll_obj_path"
SPEC_COLL_RESOURCE_KW = "spec_coll_resource"
SPEC_COLL_RESC_HIER_KW = "spec_coll_resc_hier"
SPEC_COLL_PHY_PATH_KW = "spec_coll_phy_path"
SPEC_COLL_CACHE_DIR_KW = "spec_coll_cache_dir"
SPEC_COLL_CACHE_DIRTY = "spec_coll_cache_dirty"
SPEC_COLL_REPL_NUM = "spec_coll_repl_num"


DISABLE_STRICT_ACL_KW = "disable_strict_acls"
